from abc import abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Generator, Iterator, Optional, Sequence

import numpy as np


class ValueType(str, Enum):
    """Specifies supported value types supported by Variables."""

    REAL = "real"
    SIGMOID = "sigmoid"
    PROBABILITY = "probability"  # single probability
    PROBABILITY_SAMPLE = "probability_sample"  # sample from single probability
    PROBABILITY_DISTRIBUTION = (
        "probability_distribution"  # probability distribution over classes
    )
    CLASS = "class"  # sample from probability distribution over classes


class Variable:
    """Describes an experimental variable: name, type, range, units, and value of a variable."""

    def __init__(
        self,
        name: str = "",
        value_range: Sequence = (0, 1),
        units: str = "",
        type: ValueType = ValueType.REAL,
        variable_label: str = "",
        rescale: float = 1,
        is_covariate: bool = False,
    ):
        """
        Initialize a variable.

        Args:
            name: name of the variable
            value_range: range of the variable
            units: units of the variable
            type: type of the variable
            variable_label: label of the variable
            rescale: rescale factor for the variable
            is_covariate: whether this variable is a covariate
        """

        self._name = name
        self._units = units
        self._value_range = value_range
        self._value = 0
        self.type = type
        if variable_label == "":
            self._variable_label = self._name
        else:
            self._variable_label = variable_label
        self._rescale = rescale
        self._is_covariate = is_covariate

    def __get_value_range__(self):
        """Get range of variable.
        The variable range determines the minimum and maximum allowed value
        to be manipulated or measured."""
        return self._value_range

    def __set_value_range__(self, value_range: Sequence):
        """Set range of variable.
        The variable range determines the minimum and maximum allowed value
        to be manipulated or measured.

        Arguments:
            value_range: range of the variable
        """
        self._value_range = value_range

    def __cap_value__(self, value: float):
        """Cap value of variable.

        Arguments:
            value: value of the variable
        """
        minimum = self._value_range[0]
        maximum = self._value_range[1]
        return np.min([np.max([value, minimum]), maximum])

    def get_value(self) -> float:
        """Get value.

        Returns:
            value of the variable
        """
        return self._value * self._rescale

    def set_value(self, value: float):
        """Set value.

        Arguments:
            value: value of the variable
        """
        self._value = self.__cap_value__(value)

    def get_value_from_dict(self, dictionary: dict, position: int) -> float:
        """Reads value of independent variable from a dictionary.

        Arguments:
            dictionary: dictionary with variable_label being the key
            position: position of the value in the dictionary

        Returns:
            value of the variable
        """

        value_list = dictionary[self.get_name()]

        if position > len(value_list):
            raise Exception(
                f"Queried position "
                f"{str(position)}"
                f" for variable "
                f"{self.get_name()}"
                f"'exceeds number of available positions for that variable in the dictionary."
            )

        return value_list[position] * self._rescale

    def get_value_list_from_dict(self, dictionary: dict) -> Sequence:
        """Gets the rescaled values of independent variables from a dictionary.

        Arguments:
            dictionary: dictionary with variable_label being the key

        Returns:
            list of values of the variable
        """
        value_list = dictionary[self.get_name()]

        rescaled_list = [element * self._rescale for element in value_list]

        return rescaled_list

    def set_value_from_dict(self, dictionary: dict, position: int):
        """
        Reads and sets value of independent variable from a dictionary
        with variable_label being the key

        Arguments:
            dictionary: dictionary with variable_label being the key
            position: position of the value in the dictionary
        """

        value_list = dictionary[self.get_name()]

        if position > len(value_list):
            raise Exception(
                f"Queried position "
                f"{str(position)}"
                f" for variable "
                f"{self.get_name()}"
                f" exceeds number of available positions for that variable in the dictionary."
            )

        self.set_value(value_list[position])

    def get_name(self) -> str:
        """Get variable name."""
        return self._name

    def set_name(self, name: str):
        """Set variable name.

        Arguments:
                name: name of the variable
        """
        self._name = name

    def get_units(self) -> str:
        """Get variable units.

        Returns:
            variable units
        """
        return self._units

    def set_units(self, units: str):
        """Set variable units.

        Arguments:
            units: units of the variable
        """
        self._units = units

    def get_variable_label(self) -> str:
        """Get variable label.

        Returns:
            variable label
        """
        return self._variable_label

    def set_variable_label(self, variable_label: str):
        """Set variable label.

        Arguments:
            variable_label: variable label
        """
        self._variable_label = variable_label

    def set_covariate(self, is_covariate: bool):
        """Set whether this dependent variable is treated as covariate.

        Arguments:
            is_covariate: whether this dependent variable is treated as covariate
        """
        self._is_covariate = is_covariate


@dataclass(frozen=True)
class VariableCollection:
    """Immutable metadata about dependent / independent variables and covariates."""

    independent_variables: Sequence[Variable]
    dependent_variables: Sequence[Variable]
    covariates: Sequence[Variable] = field(default_factory=list)

    @property
    def all_variables(self) -> Iterator[Variable]:
        """Get all variables.

        Returns:
            iterator over all variables
        """
        for vars in (
            self.independent_variables,
            self.dependent_variables,
            self.covariates,
        ):
            for v in vars:
                yield v

    @property
    def variable_names(self) -> Generator[str, None, None]:
        """Get variable names.

        Returns:
            variable names
        """
        return (v.get_name() for v in self.all_variables)

    @property
    def output_type(self) -> str:
        """Get output type.

        Returns:
            output type
        """
        first_type = self.dependent_variables[0].type
        assert all(dv.type == first_type for dv in self.dependent_variables), (
            "Dependent variable output types don't match. "
            "Different output types are not supported yet."
        )
        return first_type

    @property
    def input_dimensions(self) -> int:
        """The number of independent variables and covariates.

        Returns:
            number of independent variables and covariates
        """
        return len(self.independent_variables) + len(self.covariates)

    @property
    def output_dimensions(self) -> int:
        """The number of dependent variables.

        Returns:
            number of dependent variables
        """
        return len(self.dependent_variables)


class IV(Variable):
    """Independent variable."""

    def __init__(self, *args, **kwargs):
        """
        Initialize independent variable.

        For arguments, see [autora.variable.Variable][autora.variable.Variable.__init__]
        """
        self._name = "IV"
        self._variable_label = "Independent Variable"

        super().__init__(*args, **kwargs)

    # Method for measuring dependent variable.
    @abstractmethod
    def manipulate(self):
        """Manipulate independent variable."""
        pass


class DV(Variable):
    """Dependent variable."""

    def __init__(self, *args, **kwargs):
        """
        Initialize dependent variable.

        For arguments, see [autora.variable.Variable][autora.variable.Variable.__init__]
        """
        self._name = "DV"
        self._variable_label = "Dependent Variable"

        self._is_covariate = False

        super().__init__(*args, **kwargs)

    # Method for measuring dependent variable.
    @abstractmethod
    def measure(self):
        """Measure dependent variable."""
        pass

    # Get whether this dependent variable is treated as covariate.
    def __is_covariate__(self) -> bool:
        """Get whether this dependent variable is treated as covariate.

        Returns:
            whether this dependent variable is treated as covariate
        """
        return self._is_covariate

    # Set whether this dependent variable is treated as covariate.
    def __set_covariate__(self, is_covariate: bool):
        """Set whether this dependent variable is treated as covariate.

        Arguments:
            is_covariate: whether this dependent variable is treated as covariate
        """
        self._is_covariate = is_covariate


class IVTrial(IV):
    """
    Experiment trial as independent variable.
    """

    _name = "trial"
    _UID = ""
    _variable_label = "Trial"
    _units = "trials"
    _priority = 0
    _value_range = (0, 10000000)
    _value = 0

    def __init__(self, *args, **kwargs):
        """
        Initialize independent variable representing experiment trials.

        For arguments, see [autora.variable.Variable][autora.variable.Variable.__init__]
        """
        super(IVTrial, self).__init__(*args, **kwargs)

    # Waits until specified time has passed relative to reference time
    def manipulate(self):
        """
        Manipulate independent variable representing experiment trials.
        """
        pass

    def disconnect(self):
        """
        Disconnect independent variable representing experiment trials.
        """
        pass


dv_labels = {}
iv_labels = {}


def register_iv_label(**kwargs):
    iv_labels.update(kwargs)
    return


def register_dv_label(**kwargs):
    dv_labels.update(kwargs)
    return



import random
import warnings
from enum import Enum
from typing import Callable, List, Literal, Sequence, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable

from autora.theorist.darts.fan_out import Fan_Out
from autora.theorist.darts.operations import (
    OPS,
    PRIMITIVES,
    Genotype,
    get_operation_label,
    isiterable,
)


class DARTSType(str, Enum):
    """
    Enumerator that indexes different variants of DARTS.
    """

    # Liu, Simonyan & Yang (2018). Darts: Differentiable architecture search
    ORIGINAL = "original"

    # Chu, Zhou, Zhang & Li (2020). Fair darts: Eliminating unfair advantages
    # in differentiable architecture search
    FAIR = "fair"


# for 2 input nodes, 1 output node and 4 intermediate nodes,
# there are 14 possible edges (x 8 operations)
# Let input nodes be 1, 2 intermediate nodes 3, 4, 5, 6, and output node 7
# The edges are 3-1, 3-2; 4-1, 4-2, 4-3; 5-1, 5-2, 5-3, 5-4; 6-1, 6-2,
# 6-3, 6-4, 6-5; 2 + 3 + 4 + 5 = 14 edges


class MixedOp(nn.Module):
    """
    Mixture operation as applied in Differentiable Architecture Search (DARTS).
    A mixture operation amounts to a weighted mixture of a pre-defined set of operations
    that is applied to an input variable.
    """

    def __init__(self, primitives: Sequence[str] = PRIMITIVES):
        """
        Initializes a mixture operation based on a pre-specified set of primitive operations.

        Arguments:
            primitives: list of primitives to be used in the mixture operation
        """
        super(MixedOp, self).__init__()
        self._ops = nn.ModuleList()
        # loop through all the 8 primitive operations
        for primitive in primitives:
            # OPS returns an nn module for a given primitive (defines as a string)
            op = OPS[primitive]

            # add the operation
            self._ops.append(op)

    def forward(self, x: torch.Tensor, weights: torch.Tensor) -> float:
        """
        Computes a mixture operation as a weighted sum of all primitive operations.

        Arguments:
            x: input to the mixture operations
            weights: weight vector containing the weights associated with each operation

        Returns:
            y: result of the weighted mixture operation
        """
        # there are 8 weights for all the eight primitives. then it returns the
        # weighted sum of all operations performed on a given input
        return sum(w * op(x) for w, op in zip(weights, self._ops))


# Let a cell be a DAG(directed acyclic graph) containing N nodes (2 input
# nodes 1 output node?)
class Cell(nn.Module):
    """
    A cell as defined in differentiable architecture search. A single cell corresponds
    to a computation graph with the number of input nodes defined by n_input_states and
    the number of hidden nodes defined by steps. Input nodes only project to hidden nodes and hidden
    nodes project to each other with an acyclic connectivity pattern. The output of a cell
    corresponds to the concatenation of all hidden nodes. Hidden nodes are computed by integrating
    transformed outputs from sending nodes. Outputs from sending nodes correspond to
    mixture operations, i.e. a weighted combination of pre-specified operations applied to the
    variable specified by the sending node (see MixedOp).

    Attributes:
        _steps: number of hidden nodes
        _n_input_states: number of input nodes
        _ops: list of mixture operations (amounts to the list of edges in the cell)
    """

    def __init__(
        self,
        steps: int = 2,
        n_input_states: int = 1,
        primitives: Sequence[str] = PRIMITIVES,
    ):
        """
        Initializes a cell based on the number of hidden nodes (steps)
        and the number of input nodes (n_input_states).

        Arguments:
            steps: number of hidden nodes
            n_input_states: number of input nodes
        """
        # The first and second nodes of cell k are set equal to the outputs of
        # cell k − 2 and cell k − 1, respectively, and 1 × 1 convolutions
        # (ReLUConvBN) are inserted as necessary
        super(Cell, self).__init__()

        # set parameters
        self._steps = steps  # hidden nodes
        self._n_input_states = n_input_states  # input nodes

        # EDIT 11/04/19 SM: adapting to new SimpleNet data (changed from
        # multiplier to steps)
        self._multiplier = steps

        # set operations according to number of modules (empty)
        self._ops = nn.ModuleList()
        # iterate over edges: edges between each hidden node and input nodes +
        # prev hidden nodes
        for i in range(self._steps):  # hidden nodes
            for j in range(self._n_input_states + i):  # 2 refers to the 2 input nodes
                # defines the stride for link between cells
                # adds a mixed operation (derived from architecture parameters alpha)
                # for 4 intermediate nodes, a total of 14 connections
                # (MixedOps) is added
                op = MixedOp(primitives)
                # appends cell with mixed operation
                self._ops.append(op)

    def forward(self, input_states: List, weights: torch.Tensor):
        """
        Computes the output of a cell given a list of input states
        (variables represented in input nodes) and a weight matrix specifying the weights of each
        operation for each edge.

        Arguments:
            input_states: list of input nodes
            weights: matrix specifying architecture weights, i.e. the weights associated
                with each operation for each edge
        """
        # initialize states (activities of each node in the cell)
        states = list()

        # add each input node to the number of states
        for input in input_states:
            states.append(input)

        offset = 0
        # this computes the states from intermediate nodes and adds them to the list of states
        # (values of nodes)
        # for each hidden node, compute edge between existing states (input
        # nodes / previous hidden) nodes and current node
        for i in range(
            self._steps
        ):  # compute the state for each hidden node, first hidden node is
            # sum of input nodes, second is sum of input and first hidden
            s = sum(
                self._ops[offset + j](h, weights[offset + j])
                for j, h in enumerate(states)
            )
            offset += len(states)
            states.append(s)

        # concatenates the states of the last n (self._multiplier) intermediate
        # nodes to get the output of a cell
        result = torch.cat(states[-self._multiplier :], dim=1)
        return result


class Network(nn.Module):
    """
    A PyTorch computation graph according to DARTS.
    It consists of a single computation cell which transforms an
    input vector (containing all input variable) into an output vector, by applying a set of
    mixture operations which are defined by the architecture weights (labeled "alphas" of the
    network).

    The network flow looks as follows: An input vector (with _n_input_states elements) is split into
    _n_input_states separate input nodes (one node per element). The input nodes are then passed
    through a computation cell with _steps hidden nodes (see Cell). The output of the computation
    cell corresponds to the concatenation of its hidden nodes (a single vector). The final output
    corresponds to a (trained) affine transformation of this concatenation (labeled "classifier").

    Attributes:
        _n_input_states: length of input vector (translates to number of input nodes)
        _num_classes: length of output vector
        _criterion: optimization criterion used to define the loss
        _steps: number of hidden nodes in the cell
        _architecture_fixed: specifies whether the architecture weights shall remain fixed
            (not trained)
        _classifier_weight_decay: a weight decay applied to the classifier

    """

    def __init__(
        self,
        num_classes: int,
        criterion: Callable,
        steps: int = 2,
        n_input_states: int = 2,
        architecture_fixed: bool = False,
        train_classifier_coefficients: bool = False,
        train_classifier_bias: bool = False,
        classifier_weight_decay: float = 0,
        darts_type: DARTSType = DARTSType.ORIGINAL,
        primitives: Sequence[str] = PRIMITIVES,
    ):
        """
        Initializes the network.

        Arguments:
            num_classes: length of output vector
            criterion: optimization criterion used to define the loss
            steps: number of hidden nodes in the cell
            n_input_states: length of input vector (translates to number of input nodes)
            architecture_fixed: specifies whether the architecture weights shall remain fixed
            train_classifier_coefficients: specifies whether the classifier coefficients shall be
                trained
            train_classifier_bias: specifies whether the classifier bias shall be trained
            classifier_weight_decay: a weight decay applied to the classifier
            darts_type: variant of DARTS (regular or fair) that is applied for training
        """
        super(Network, self).__init__()

        # set parameters
        self._num_classes = num_classes  # number of output classes
        self._criterion = criterion  # optimization criterion (e.g., softmax)
        self._steps = steps  # the number of intermediate nodes (e.g., 2)
        self._n_input_states = n_input_states  # number of input nodes
        self.DARTS_type = darts_type  # darts variant
        self._multiplier = (
            1  # the number of internal nodes that get concatenated to the output
        )
        self.primitives = primitives

        # set parameters
        self._dim_output = self._steps
        self._architecture_fixed = architecture_fixed
        self._classifier_weight_decay = classifier_weight_decay

        # input nodes
        self.stem = nn.Sequential(Fan_Out(self._n_input_states))

        self.cells = (
            nn.ModuleList()
        )  # get list of all current modules (should be empty)

        # generate a cell that undergoes architecture search
        self.cells = Cell(steps, self._n_input_states, self.primitives)

        # last layer is a linear classifier (e.g. with 10 CIFAR classes)
        self.classifier = nn.Linear(
            self._dim_output, num_classes
        )  # make this the number of input states

        # initialize classifier weights
        if train_classifier_coefficients is False:
            self.classifier.weight.data.fill_(1)
            self.classifier.weight.requires_grad = False

        if train_classifier_bias is False:
            self.classifier.bias.data.fill_(0)
            self.classifier.bias.requires_grad = False

        # initializes weights of the architecture
        self._initialize_alphas()

    # function for copying the network
    def new(self) -> nn.Module:
        """
        Returns a copy of the network.

        Returns:
            a copy of the network

        """

        model_new = Network(
            # self._C, self._num_classes, self._criterion, steps=self._steps
            num_classes=self._num_classes,
            criterion=self._criterion,
            steps=self._steps,
            n_input_states=self._n_input_states,
            architecture_fixed=self._architecture_fixed,
            classifier_weight_decay=self._classifier_weight_decay,
            darts_type=self.DARTS_type,
            primitives=self.primitives,
        )

        for x, y in zip(model_new.arch_parameters(), self.arch_parameters()):
            x.data.copy_(y.data)
        return model_new

    # computes forward pass for full network
    def forward(self, x: torch.Tensor):
        """
        Computes output of the network.

        Arguments:
            x: input to the network
        """

        # compute stem first
        input_states = self.stem(x)

        # get architecture weights
        if self._architecture_fixed:
            weights = self.alphas_normal
        else:
            if self.DARTS_type == DARTSType.ORIGINAL:
                weights = F.softmax(self.alphas_normal, dim=-1)
            elif self.DARTS_type == DARTSType.FAIR:
                weights = torch.sigmoid(self.alphas_normal)
            else:
                raise Exception(
                    "DARTS Type " + str(self.DARTS_type) + " not implemented"
                )

        # then apply cell with weights
        cell_output = self.cells(input_states, weights)

        # compute logits
        logits = self.classifier(cell_output.view(cell_output.size(0), -1))
        # just gets output to have only 2 dimensions (batch_size x num units in
        # output layer)

        return logits

    def _loss(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Computes the loss of the network for the specified criterion.

        Arguments:
            input: input patterns
            target: target patterns

        Returns:
            loss
        """
        logits = self(input)
        return self._criterion(logits, target)  # returns cross entropy by default

    # regularization
    def apply_weight_decay_to_classifier(self, lr: float):
        """
        Applies a weight decay to the weights projecting from the cell to the final output layer.

        Arguments:
            lr: learning rate
        """
        # weight decay proportional to degrees of freedom
        for p in self.classifier.parameters():
            if p.requires_grad is False:
                continue
            p.data.sub_(
                self._classifier_weight_decay
                * lr
                * torch.sign(p.data)
                * (torch.abs(p.data))
            )  # weight decay

    def _initialize_alphas(self):
        """
        Initializes the architecture weights.
        """
        # compute the number of possible connections between nodes
        k = sum(1 for i in range(self._steps) for n in range(self._n_input_states + i))
        # number of available primitive operations (8 different types for a
        # conv net)
        num_ops = len(self.primitives)

        # e.g., generate 14 (number of available edges) by 8 (operations)
        # weight matrix for normal alphas of the architecture
        self.alphas_normal = Variable(
            1e-3 * torch.randn(k, num_ops), requires_grad=True
        )
        # those are all the parameters of the architecture
        self._arch_parameters = [self.alphas_normal]

    # provide back the architecture as a parameter
    def arch_parameters(self) -> List:
        """
        Returns architecture weights.

        Returns:
            _arch_parameters: architecture weights.
        """
        return self._arch_parameters

    # fixes architecture
    def fix_architecture(self, switch: bool, new_weights: torch.Tensor = None):
        """
        Freezes or unfreezes the architecture weights.

        Arguments:
            switch: set true to freeze architecture weights or false unfreeze
            new_weights: new set of architecture weights
        """
        self._architecture_fixed = switch
        if new_weights is not None:
            self.alphas_normal = new_weights
        return

    def sample_alphas_normal(
        self, sample_amp: float = 1, fair_darts_weight_threshold: float = 0
    ) -> torch.Tensor:
        """
        Samples an architecture from the mixed operations from a probability distribution that is
        defined by the (softmaxed) architecture weights.
        This amounts to selecting one operation per edge (i.e., setting the architecture
        weight of that operation to one while setting the others to zero).

        Arguments:
            sample_amp: temperature that is applied before passing the weights through a softmax
            fair_darts_weight_threshold: used in fair DARTS. If an architecture weight is below
                this value then it is set to zero.

        Returns:
            alphas_normal_sample: sampled architecture weights.
        """

        alphas_normal = self.alphas_normal.clone()
        alphas_normal_sample = Variable(torch.zeros(alphas_normal.data.shape))

        for edge in range(alphas_normal.data.shape[0]):
            if self.DARTS_type == DARTSType.ORIGINAL:
                W_soft = F.softmax(alphas_normal[edge] * sample_amp, dim=0)
            elif self.DARTS_type == DARTSType.FAIR:
                transformed_alphas_normal = alphas_normal[edge]
                above_threshold = False
                for idx in range(len(transformed_alphas_normal.data)):
                    if (
                        torch.sigmoid(transformed_alphas_normal).data[idx]
                        > fair_darts_weight_threshold
                    ):
                        above_threshold = True
                        break
                if above_threshold:
                    W_soft = F.softmax(transformed_alphas_normal * sample_amp, dim=0)
                else:
                    W_soft = Variable(torch.zeros(alphas_normal[edge].shape))
                    W_soft[self.primitives.index("none")] = 1

            else:
                raise Exception(
                    "DARTS Type " + str(self.DARTS_type) + " not implemented"
                )

            if torch.any(W_soft != W_soft):
                warnings.warn(
                    "Cannot properly sample from architecture weights due to nan entries."
                )
                k_sample = random.randrange(len(W_soft))
            else:
                k_sample = np.random.choice(range(len(W_soft)), p=W_soft.data.numpy())
            alphas_normal_sample[edge, k_sample] = 1

        return alphas_normal_sample

    def max_alphas_normal(self) -> torch.Tensor:
        """
        Samples an architecture from the mixed operations by selecting, for each edge,
        the operation with the largest architecture weight.

        Returns:
            alphas_normal_sample: sampled architecture weights.
        """
        alphas_normal = self.alphas_normal.clone()
        alphas_normal_sample = Variable(torch.zeros(alphas_normal.data.shape))

        for edge in range(alphas_normal.data.shape[0]):
            row = alphas_normal[edge]
            max_idx = np.argmax(row.data)
            alphas_normal_sample[edge, max_idx] = 1

        return alphas_normal_sample

    # returns the genotype of the model
    def genotype(self, sample: bool = False) -> Genotype:
        """
        Computes a genotype of the model which specifies the current computation graph based on
        the largest architecture weight for each edge, or based on a sample.
        The genotype can be used for parsing or plotting the computation graph.

        Arguments:
            sample: if set to true, the architecture will be determined by sampling
                from a probability distribution that is determined by the
                softmaxed architecture weights. If set to false (default), the architecture will be
                determined based on the largest architecture weight per edge.

        Returns:
            genotype: genotype describing the current (sampled) architecture
        """
        # this function uses the architecture weights to retrieve the
        # operations with the highest weights
        def _parse(weights):
            gene = []
            n = (
                self._n_input_states
            )  # 2 ... changed this to adapt to number of input states
            start = 0
            for i in range(self._steps):
                end = start + n
                W = weights[start:end].copy()
                # first get all the edges for a given node, edges are sorted according to their
                # highest (non-none) weight, starting from the edge with the smallest heighest
                # weight

                if "none" in self.primitives:
                    none_index = self.primitives.index("none")
                else:
                    none_index = -1

                edges = sorted(
                    range(n),
                    key=lambda x: -max(
                        W[x][k] for k in range(len(W[x])) if k != none_index
                    ),
                )
                # for each edge, figure out which is the primitive with the
                # highest
                for (
                    j
                ) in edges:  # looping through all the edges for the current node (i)
                    if sample:
                        W_soft = F.softmax(Variable(torch.from_numpy(W[j])))
                        k_best = np.random.choice(
                            range(len(W[j])), p=W_soft.data.numpy()
                        )
                    else:
                        k_best = None
                        # looping through all the primitives
                        for k in range(len(W[j])):
                            # choose the primitive with the highest weight
                            # if k != self.primitives.index('none'):
                            # EDIT SM 01/13: commented to include "none"
                            # weights in genotype
                            if k_best is None or W[j][k] > W[j][k_best]:
                                k_best = k
                        # add gene (primitive, edge number)
                    gene.append((self.primitives[k_best], j))
                start = end
                n += 1
            return gene

        if self._architecture_fixed:
            gene_normal = _parse(self.alphas_normal.data.cpu().numpy())
        else:
            gene_normal = _parse(
                F.softmax(self.alphas_normal, dim=-1).data.cpu().numpy()
            )

        concat = range(2 + self._steps - self._multiplier, self._steps + 2)
        genotype = Genotype(
            normal=gene_normal,
            normal_concat=concat,
        )
        return genotype

    def count_parameters(self, print_parameters: bool = False) -> Tuple[int, int, list]:
        """
        Counts and returns the parameters (coefficients) of the architecture defined by the
        highest architecture weights.

        Arguments:
            print_parameters: if set to true, the function will print all parameters.

        Returns:
            n_params_total: total number of parameters
            n_params_base: number of parameters determined by the classifier
            param_list: list of parameters specifying the corresponding edge (operation)
                and value
        """

        # counts only parameters of operations with the highest architecture weight
        n_params_total = 0

        # count classifier
        for parameter in self.classifier.parameters():
            if parameter.requires_grad is True:
                n_params_total += parameter.data.numel()

        # count stem
        for parameter in self.stem.parameters():
            if parameter.requires_grad is True:
                n_params_total += parameter.data.numel()

        n_params_base = (
            n_params_total  # number of parameters, excluding individual cells
        )

        param_list = list()
        # now count number of parameters for cells that have highest
        # probability
        for idx, op in enumerate(self.cells._ops):
            # pick most operation with highest likelihood
            values = self.alphas_normal[idx, :].data.numpy()
            maxIdx = np.where(values == max(values))

            tmp_param_list = list()
            if isiterable(op._ops[maxIdx[0].item(0)]):  # Zero is not iterable

                for subop in op._ops[maxIdx[0].item(0)]:

                    for parameter in subop.parameters():
                        tmp_param_list.append(parameter.data.numpy().squeeze())
                        if parameter.requires_grad is True:
                            n_params_total += parameter.data.numel()

            if print_parameters:
                print(
                    "Edge ("
                    + str(idx)
                    + "): "
                    + get_operation_label(
                        self.primitives[maxIdx[0].item(0)], tmp_param_list
                    )
                )
            param_list.append(tmp_param_list)

        # # get parameters from final linear classifier
        # tmp_param_list = list()
        # for parameter in self.classifier.parameters():
        #   for subparameter in parameter:
        #     tmp_param_list.append(subparameter.data.numpy().squeeze())

        # get parameters from final linear for each edge
        for edge in range(self._steps):
            tmp_param_list = list()
            # add weight
            tmp_param_list.append(
                self.classifier._parameters["weight"].data[:, edge].numpy()
            )
            # add partial bias (bias of classifier units will be divided by
            # number of edges)
            if "bias" in self.classifier._parameters.keys() and edge == 0:
                tmp_param_list.append(self.classifier._parameters["bias"].data.numpy())
            param_list.append(tmp_param_list)

            if print_parameters:
                print(
                    "Classifier from Node "
                    + str(edge)
                    + ": "
                    + get_operation_label("classifier_concat", tmp_param_list)
                )

        return (n_params_total, n_params_base, param_list)

    def architecture_to_str_list(
        self,
        input_labels: Sequence[str],
        output_labels: Sequence[str],
        output_function_label: str = "",
        decimals_to_display: int = 2,
        output_format: Literal["latex", "console"] = "console",
    ) -> List:
        """
        Returns a list of strings representing the model.

        Arguments:
            input_labels: list of strings representing the input states.
            output_labels: list of strings representing the output states.
            output_function_label: string representing the output function.
            decimals_to_display: number of decimals to display.
            output_format: if set to `"console"`, returns equations formatted for the command line,
                if set to `"latex"`, returns equations in latex format


        Returns:
            list of strings representing the model
        """
        (n_params_total, n_params_base, param_list) = self.count_parameters(
            print_parameters=False
        )
        genotype = self.genotype().normal
        steps = self._steps
        edge_list = list()

        n = len(input_labels)
        start = 0
        for i in range(steps):  # for every node
            end = start + n
            # for k in [2*i, 2*i + 1]:

            edge_operations_list = list()
            op_list = list()

            for k in range(start, end):
                if (
                    output_format == "latex"
                ):  # for every edge projecting to current node
                    v = "k_" + str(i + 1)
                else:
                    v = "k" + str(i + 1)
                op, j = genotype[k]
                if j < len(input_labels):
                    u = input_labels[j]
                else:
                    if output_format == "latex":
                        u = "k_" + str(j - len(input_labels) + 1)
                    else:
                        u = "k" + str(j - len(input_labels) + 1)
                if op != "none":
                    op_label = op
                    params = param_list[
                        start + j
                    ]  # note: genotype order and param list order don't align
                    op_label = get_operation_label(
                        op,
                        params,
                        decimals=decimals_to_display,
                        input_var=u,
                        output_format=output_format,
                    )
                    op_list.append(op)
                    edge_operations_list.append(op_label)

            if len(edge_operations_list) == 0:
                edge_str = v + " = 0"
            else:
                edge_str = ""
            for i, edge_operation in enumerate(edge_operations_list):
                if i == 0:
                    edge_str += v + " = " + edge_operation
                if i > 0:
                    if (
                        op_list[i] != "add"
                        and op_list[i] != "subtract"
                        and op_list[i] != "none"
                    ):
                        edge_str += " +"
                    edge_str += " " + edge_operation

            edge_list.append(edge_str)
            start = end
            n += 1

        # TODO: extend to multiple outputs
        if output_format == "latex":
            classifier_str = output_labels[0] + " = " + output_function_label
            if output_function_label != "":
                classifier_str += "\\left("
        else:
            classifier_str = output_labels[0] + " = " + output_function_label
            if output_function_label != "":
                classifier_str += "("

        bias = None
        for i in range(steps):
            param_idx = len(param_list) - steps + i
            tmp_param_list = param_list[param_idx]
            if i == 0 and len(tmp_param_list) == 2:
                bias = tmp_param_list[1]
            if i > 0:
                classifier_str += " + "

            if output_format == "latex":
                input_var = "k_" + str(i + 1)
            else:
                input_var = "k" + str(i + 1)

            classifier_str += get_operation_label(
                "classifier",
                tmp_param_list[0],
                decimals=decimals_to_display,
                input_var=input_var,
            )

            if i == steps - 1 and bias is not None:
                classifier_str += " + " + str(bias[0])

            if i == steps - 1:
                if output_function_label != "":
                    if output_format == "latex":
                        classifier_str += "\\right)"
                    else:
                        classifier_str += ")"

        edge_list.append(classifier_str)

        return edge_list



import copy
import logging
from dataclasses import dataclass
from itertools import cycle
from types import SimpleNamespace
from typing import Any, Callable, Iterator, Literal, Optional, Sequence, Tuple

import numpy as np
import torch
import torch.nn
import torch.nn.utils
import torch.utils.data
import tqdm
from matplotlib import pyplot as plt
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y

from autora.theorist.darts import (
    PRIMITIVES,
    Architect,
    AvgrageMeter,
    DARTSType,
    Network,
    darts_dataset_from_ndarray,
    darts_model_plot,
    format_input_target,
    get_loss_function,
    get_output_format,
    get_output_str,
)
from autora.variable import ValueType

_logger = logging.getLogger(__name__)

_progress_indicator = tqdm.auto.tqdm

SAMPLING_STRATEGIES = Literal["max", "sample"]
IMPLEMENTED_DARTS_TYPES = Literal["original", "fair"]
IMPLEMENTED_OUTPUT_TYPES = Literal[
    "real",
    "sigmoid",
    "probability",
    "probability_sample",
    "probability_distribution",
]


@dataclass(frozen=True)
class _DARTSResult:
    """A container for passing fitted DARTS results around."""

    network: Network
    model: torch.nn.Module


def _general_darts(
    X: np.ndarray,
    y: np.ndarray,
    network: Optional[Network] = None,
    batch_size: int = 20,
    num_graph_nodes: int = 2,
    output_type: IMPLEMENTED_OUTPUT_TYPES = "real",
    classifier_weight_decay: float = 1e-2,
    darts_type: IMPLEMENTED_DARTS_TYPES = "original",
    init_weights_function: Optional[Callable] = None,
    param_updates_per_epoch: int = 20,
    param_updates_for_sampled_model: int = 100,
    param_learning_rate_max: float = 2.5e-2,
    param_learning_rate_min: float = 0.01,
    param_momentum: float = 9e-1,
    param_weight_decay: float = 3e-4,
    arch_learning_rate_max: float = 3e-3,
    arch_updates_per_epoch: int = 20,
    arch_weight_decay: float = 1e-4,
    arch_weight_decay_df: float = 3e-4,
    arch_weight_decay_base: float = 0.0,
    arch_momentum: float = 9e-1,
    fair_darts_loss_weight: int = 1,
    max_epochs: int = 100,
    grad_clip: float = 5,
    primitives: Sequence[str] = PRIMITIVES,
    train_classifier_coefficients: bool = False,
    train_classifier_bias: bool = False,
    execution_monitor: Callable = (lambda *args, **kwargs: None),
    sampling_strategy: SAMPLING_STRATEGIES = "max",
) -> _DARTSResult:
    """
    Function to implement the DARTS optimization, given a fixed architecture and input data.

    Arguments:
        X: Input data.
        y: Target data.
        batch_size: Batch size for the data loader.
        num_graph_nodes: Number of nodes in the desired computation graph.
        output_type: Type of output function to use. This function is applied to transform
        the output of the mixture architecture.
        classifier_weight_decay: Weight decay for the classifier.
        darts_type: Type of DARTS to use ('original' or 'fair').
        init_weights_function: Function to initialize the parameters of each operation.
        param_learning_rate_max: Initial (maximum) learning rate for the operation parameters.
        param_learning_rate_min: Final (minimum) learning rate for the operation parameters.
        param_momentum: Momentum for the operation parameters.
        param_weight_decay: Weight decay for the operation parameters.
        param_updates_per_epoch: Number of updates to perform per epoch.
        for the operation parameters.
        arch_learning_rate_max: Initial (maximum) learning rate for the architecture.
        arch_updates_per_epoch: Number of architecture weight updates to perform per epoch.
        arch_weight_decay: Weight decay for the architecture weights.
        arch_weight_decay_df: An additional weight decay that scales with the number of parameters
        (degrees of freedom) in the operation. The higher this weight decay, the more DARTS will
        prefer simple operations.
        arch_weight_decay_base: A base weight decay that is added to the scaled weight decay.
        arch_momentum: Momentum for the architecture weights.
        fair_darts_loss_weight: Weight of the loss in fair darts which forces architecture weights
        to become either 0 or 1.
        max_epochs: Maximum number of epochs to train for.
        grad_clip: Gradient clipping value for updating the parameters of the operations.
        primitives: List of primitives (operations) to use.
        train_classifier_coefficients: Whether to train the coefficients of the classifier.
        train_classifier_bias: Whether to train the bias of the classifier.
        execution_monitor: Function to monitor the execution of the model.

    Returns:
        A _DARTSResult object containing the fitted model and the network architecture.
    """

    _logger.info("Starting fit initialization")

    data_loader, input_dimensions, output_dimensions = _get_data_loader(
        X=X,
        y=y,
        batch_size=batch_size,
    )

    criterion = get_loss_function(ValueType(output_type))
    output_function = get_output_format(ValueType(output_type))

    if network is None:
        network = Network(
            num_classes=output_dimensions,
            criterion=criterion,
            steps=num_graph_nodes,
            n_input_states=input_dimensions,
            classifier_weight_decay=classifier_weight_decay,
            darts_type=DARTSType(darts_type),
            primitives=primitives,
            train_classifier_coefficients=train_classifier_coefficients,
            train_classifier_bias=train_classifier_bias,
        )

    if init_weights_function is not None:
        network.apply(init_weights_function)

    # Generate the architecture of the model
    architect = Architect(
        network,
        arch_momentum=arch_momentum,
        arch_weight_decay=arch_weight_decay,
        arch_weight_decay_df=arch_weight_decay_df,
        arch_weight_decay_base=arch_weight_decay_base,
        fair_darts_loss_weight=fair_darts_loss_weight,
        arch_learning_rate_max=arch_learning_rate_max,
    )

    _logger.info("Starting fit.")
    network.train()

    for epoch in _progress_indicator(range(max_epochs)):

        _logger.debug(f"Running fit, epoch {epoch}")

        data_iterator = _get_data_iterator(data_loader)

        # Do the Architecture update
        for arch_step in range(arch_updates_per_epoch):
            _logger.debug(
                f"Running architecture update, "
                f"epoch: {epoch}, architecture: {arch_step}"
            )

            X_batch, y_batch = _get_next_input_target(
                data_iterator, criterion=criterion
            )

            architect.step(
                input_valid=X_batch,
                target_valid=y_batch,
                network_optimizer=architect.optimizer,
                unrolled=False,
            )

        # Then run the param optimization
        _optimize_coefficients(
            network=network,
            criterion=criterion,
            data_loader=data_loader,
            grad_clip=grad_clip,
            param_learning_rate_max=param_learning_rate_max,
            param_learning_rate_min=param_learning_rate_min,
            param_momentum=param_momentum,
            param_update_steps=param_updates_per_epoch,
            param_weight_decay=param_weight_decay,
        )

        execution_monitor(**locals())

    model = _generate_model(
        network_=network,
        output_type=output_type,
        sampling_strategy=sampling_strategy,
        data_loader=data_loader,
        param_update_steps=param_updates_for_sampled_model,
        param_learning_rate_max=param_learning_rate_max,
        param_learning_rate_min=param_learning_rate_min,
        param_momentum=param_momentum,
        param_weight_decay=param_weight_decay,
        grad_clip=grad_clip,
    )

    results = _DARTSResult(model=model, network=network)

    return results


def _optimize_coefficients(
    network: Network,
    criterion: torch.nn.Module,
    data_loader: torch.utils.data.DataLoader,
    grad_clip: float,
    param_learning_rate_max: float,
    param_learning_rate_min: float,
    param_momentum: float,
    param_update_steps: int,
    param_weight_decay: float,
):
    """
    Function to optimize the coefficients of a DARTS Network.

    Warning: This modifies the coefficients of the Network in place.

    Arguments:
        network: The DARTS Network to optimize the coefficients of.
        criterion: The loss function to use.
        data_loader: The data loader to use for the optimization.
        grad_clip: Whether to clip the gradients.
        param_update_steps: The number of parameter update steps to perform.
        param_learning_rate_max: Initial (maximum) learning rate for the operation parameters.
        param_learning_rate_min: Final (minimum) learning rate for the operation parameters.
        param_momentum: Momentum for the operation parameters.
        param_weight_decay: Weight decay for the operation parameters.
    """
    optimizer = torch.optim.SGD(
        params=network.parameters(),
        lr=param_learning_rate_max,
        momentum=param_momentum,
        weight_decay=param_weight_decay,
    )
    scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
        optimizer=optimizer,
        T_max=param_update_steps,
        eta_min=param_learning_rate_min,
    )

    data_iterator = _get_data_iterator(data_loader)

    objs = AvgrageMeter()

    if network.count_parameters()[0] == 0:
        return

    for param_step in range(param_update_steps):
        _logger.debug(f"Running parameter update, " f"param: {param_step}")

        lr = scheduler.get_last_lr()[0]
        X_batch, y_batch = _get_next_input_target(data_iterator, criterion=criterion)
        optimizer.zero_grad()

        # compute loss for the model
        logits = network(X_batch)
        loss = criterion(logits, y_batch)

        # update gradients for model
        loss.backward()

        # clips the gradient norm
        torch.nn.utils.clip_grad_norm_(network.parameters(), grad_clip)

        # moves optimizer one step (applies gradients to weights)
        optimizer.step()

        # applies weight decay to classifier weights
        network.apply_weight_decay_to_classifier(lr)

        # moves the annealing scheduler forward to determine new learning rate
        scheduler.step()

        # compute accuracy metrics
        n = X_batch.size(0)
        objs.update(loss.data, n)


def _get_data_loader(
    X: np.ndarray,
    y: np.ndarray,
    batch_size: int,
) -> torch.utils.data.DataLoader:
    """Construct a minimal torch.utils.data.DataLoader for the input data.

    Arguments:
        X: The input data.
        y: The target data.
        batch_size: The batch size to use.

    Returns:
        A torch.utils.data.DataLoader for the input data.
    """

    X_, y_ = check_X_y(X, y, ensure_2d=True, multi_output=True)

    if y_.ndim == 1:
        y_ = y_.reshape((y_.size, 1))

    input_dimensions = X_.shape[1]
    output_dimensions = y_.shape[1]

    experimental_data = darts_dataset_from_ndarray(X_, y_)

    data_loader = torch.utils.data.DataLoader(
        experimental_data,
        batch_size=batch_size,
        shuffle=True,
        pin_memory=True,
        num_workers=0,
    )
    return data_loader, input_dimensions, output_dimensions


def _get_data_iterator(data_loader: torch.utils.data.DataLoader) -> Iterator:
    """Get an iterator for the data loader.

    Arguments:
        data_loader: The data loader to get the iterator for.

    Returns:
        An iterator for the data loader.
    """
    data_iterator = cycle(iter(data_loader))
    return data_iterator


def _get_next_input_target(
    data_iterator: Iterator, criterion: torch.nn.Module
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Get the next input and target from the data iterator.
    Args:
        data_iterator: The data iterator to get the next input and target from.
        criterion: The loss function to use.

    Returns:
        The next input and target from the data iterator.

    """
    input_search, target_search = next(data_iterator)

    input_var = torch.autograd.Variable(input_search, requires_grad=False)
    target_var = torch.autograd.Variable(target_search, requires_grad=False)

    input_fmt, target_fmt = format_input_target(
        input_var, target_var, criterion=criterion
    )
    return input_fmt, target_fmt


def _generate_model(
    network_: Network,
    output_type: IMPLEMENTED_OUTPUT_TYPES,
    sampling_strategy: SAMPLING_STRATEGIES,
    data_loader: torch.utils.data.DataLoader,
    param_update_steps: int,
    param_learning_rate_max: float,
    param_learning_rate_min: float,
    param_momentum: float,
    param_weight_decay: float,
    grad_clip: float,
) -> Network:
    """
    Generate a model architecture from mixed DARTS model.

    Arguments:
        sampling_strategy: The sampling strategy used to pick the operations
        based on the trained architecture weights (e.g. "max", "sample").
        network: The mixed DARTS model.
        coefficient_optimizer: The function to optimize the coefficients of the trained model
        output_type: The output value type that is used for the output of the sampled model.
        param_update_steps: The number of parameter update steps to perform.
        param_learning_rate_max: Initial (maximum) learning rate for the operation parameters.
        param_learning_rate_min: Final (minimum) learning rate for the operation parameters.
        param_momentum: Momentum for the operation parameters.
        param_weight_decay: Weight decay for the operation parameters.

    Returns:
        A model architecture that is a combination of the trained model and the output function.
    """
    criterion = get_loss_function(ValueType(output_type))
    output_function = get_output_format(ValueType(output_type))

    # Set edges in the network with the highest weights to 1, others to 0
    model_without_output_function = copy.deepcopy(network_)

    if sampling_strategy == "max":
        new_weights = model_without_output_function.max_alphas_normal()
    elif sampling_strategy == "sample":
        new_weights = model_without_output_function.sample_alphas_normal()

    model_without_output_function.fix_architecture(True, new_weights=new_weights)

    # Re-optimize the parameters

    _optimize_coefficients(
        model_without_output_function,
        criterion=criterion,
        data_loader=data_loader,
        grad_clip=grad_clip,
        param_learning_rate_max=param_learning_rate_max,
        param_learning_rate_min=param_learning_rate_min,
        param_momentum=param_momentum,
        param_update_steps=param_update_steps,
        param_weight_decay=param_weight_decay,
    )

    # Include the output function
    model = torch.nn.Sequential(model_without_output_function, output_function)

    return model


class DARTSRegressor(BaseEstimator, RegressorMixin):
    """
    Differentiable ARchiTecture Search Regressor.

    DARTS finds a composition of functions and coefficients to minimize a loss function suitable for
    the dependent variable.

    This class is intended to be compatible with the
    [Scikit-Learn Estimator API](https://scikit-learn.org/stable/developers/develop.html).

    Examples:

        >>> import numpy as np
        >>> num_samples = 1000
        >>> X = np.linspace(start=0, stop=1, num=num_samples).reshape(-1, 1)
        >>> y = 15. * np.ones(num_samples)
        >>> estimator = DARTSRegressor(num_graph_nodes=1)
        >>> estimator = estimator.fit(X, y)
        >>> estimator.predict([[0.5]])
        array([[15.051043]], dtype=float32)


    Attributes:
        network_: represents the optimized network for the architecture search, without the
            output function
        model_: represents the best-fit model including the output function
            after sampling of the network to pick a single computation graph.
            By default, this is the computation graph with the maximum weights,
            but can be set to a graph based on a sample on the edge weights
            by running the `resample_model(sample_strategy="sample")` method.
            It can be reset by running the `resample_model(sample_strategy="max")` method.



    """

    def __init__(
        self,
        batch_size: int = 64,
        num_graph_nodes: int = 2,
        output_type: IMPLEMENTED_OUTPUT_TYPES = "real",
        classifier_weight_decay: float = 1e-2,
        darts_type: IMPLEMENTED_DARTS_TYPES = "original",
        init_weights_function: Optional[Callable] = None,
        param_updates_per_epoch: int = 10,
        param_updates_for_sampled_model: int = 100,
        param_learning_rate_max: float = 2.5e-2,
        param_learning_rate_min: float = 0.01,
        param_momentum: float = 9e-1,
        param_weight_decay: float = 3e-4,
        arch_updates_per_epoch: int = 1,
        arch_learning_rate_max: float = 3e-3,
        arch_weight_decay: float = 1e-4,
        arch_weight_decay_df: float = 3e-4,
        arch_weight_decay_base: float = 0.0,
        arch_momentum: float = 9e-1,
        fair_darts_loss_weight: int = 1,
        max_epochs: int = 10,
        grad_clip: float = 5,
        primitives: Sequence[str] = PRIMITIVES,
        train_classifier_coefficients: bool = False,
        train_classifier_bias: bool = False,
        execution_monitor: Callable = (lambda *args, **kwargs: None),
        sampling_strategy: SAMPLING_STRATEGIES = "max",
    ) -> None:
        """
        Initializes the DARTSRegressor.

        Arguments:
            batch_size: Batch size for the data loader.
            num_graph_nodes: Number of nodes in the desired computation graph.
            output_type: Type of output function to use. This function is applied to transform
                the output of the mixture architecture.
            classifier_weight_decay: Weight decay for the classifier.
            darts_type: Type of DARTS to use ('original' or 'fair').
            init_weights_function: Function to initialize the parameters of each operation.
            param_updates_per_epoch: Number of updates to perform per epoch.
                for the operation parameters.
            param_learning_rate_max: Initial (maximum) learning rate for the operation parameters.
            param_learning_rate_min: Final (minimum) learning rate for the operation parameters.
            param_momentum: Momentum for the operation parameters.
            param_weight_decay: Weight decay for the operation parameters.
            arch_updates_per_epoch: Number of architecture weight updates to perform per epoch.
            arch_learning_rate_max: Initial (maximum) learning rate for the architecture.
            arch_weight_decay: Weight decay for the architecture weights.
            arch_weight_decay_df: An additional weight decay that scales with the number of
                parameters (degrees of freedom) in the operation. The higher this weight decay,
                the more DARTS will prefer simple operations.
            arch_weight_decay_base: A base weight decay that is added to the scaled weight decay.
                arch_momentum: Momentum for the architecture weights.
            fair_darts_loss_weight: Weight of the loss in fair darts which forces architecture
                weights to become either 0 or 1.
            max_epochs: Maximum number of epochs to train for.
            grad_clip: Gradient clipping value for updating the parameters of the operations.
            primitives: List of primitives (operations) to use.
            train_classifier_coefficients: Whether to train the coefficients of the classifier.
            train_classifier_bias: Whether to train the bias of the classifier.
            execution_monitor: Function to monitor the execution of the model.
            primitives: list of primitive operations used in the DARTS network,
                e.g., 'add', 'subtract', 'none'. For details, see
                [`autora.theorist.darts.operations`][autora.theorist.darts.operations]
        """

        self.batch_size = batch_size

        self.num_graph_nodes = num_graph_nodes
        self.classifier_weight_decay = classifier_weight_decay
        self.darts_type = darts_type
        self.init_weights_function = init_weights_function

        self.param_updates_per_epoch = param_updates_per_epoch
        self.param_updates_for_sampled_model = param_updates_for_sampled_model

        self.param_learning_rate_max = param_learning_rate_max
        self.param_learning_rate_min = param_learning_rate_min
        self.param_momentum = param_momentum
        self.arch_momentum = arch_momentum
        self.param_weight_decay = param_weight_decay

        self.arch_updates_per_epoch = arch_updates_per_epoch
        self.arch_weight_decay = arch_weight_decay
        self.arch_weight_decay_df = arch_weight_decay_df
        self.arch_weight_decay_base = arch_weight_decay_base
        self.arch_learning_rate_max = arch_learning_rate_max
        self.fair_darts_loss_weight = fair_darts_loss_weight

        self.max_epochs = max_epochs
        self.grad_clip = grad_clip

        self.primitives = primitives

        self.output_type = output_type
        self.darts_type = darts_type

        self.X_: Optional[np.ndarray] = None
        self.y_: Optional[np.ndarray] = None
        self.network_: Optional[Network] = None
        self.model_: Optional[Network] = None

        self.train_classifier_coefficients = train_classifier_coefficients
        self.train_classifier_bias = train_classifier_bias

        self.execution_monitor = execution_monitor

        self.sampling_strategy = sampling_strategy

    def fit(self, X: np.ndarray, y: np.ndarray):
        """
        Runs the optimization for a given set of `X`s and `y`s.

        Arguments:
            X: independent variables in an n-dimensional array
            y: dependent variables in an n-dimensional array

        Returns:
            self (DARTSRegressor): the fitted estimator
        """

        if self.output_type == "class":
            raise NotImplementedError(
                "Classification not implemented for DARTSRegressor."
            )

        params = self.get_params()

        fit_results = _general_darts(X=X, y=y, network=self.network_, **params)
        self.X_ = X
        self.y_ = y
        self.network_ = fit_results.network
        self.model_ = fit_results.model
        return self

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Applies the fitted model to a set of independent variables `X`,
        to give predictions for the dependent variable `y`.

        Arguments:
            X: independent variables in an n-dimensional array

        Returns:
            y: predicted dependent variable values
        """
        X_ = check_array(X)

        # First run the checks using the scikit-learn API, listing the key parameters
        check_is_fitted(self, attributes=["model_"])

        # Since self.model_ is initialized as None, mypy throws an error if we
        # just call self.model_(X) in the predict method, as it could still be none.
        # MyPy doesn't understand that the sklearn check_is_fitted function
        # ensures the self.model_ parameter is initialized and otherwise throws an error,
        # so we check that explicitly here and pass the model which can't be None.
        assert self.model_ is not None

        y_ = self.model_(torch.as_tensor(X_).float())
        y = y_.detach().numpy()

        return y

    def visualize_model(
        self,
        input_labels: Optional[Sequence[str]] = None,
    ):
        """
        Visualizes the model architecture as a graph.

        Arguments:
            input_labels: labels for the input nodes

        """

        check_is_fitted(self, attributes=["model_"])
        assert self.model_ is not None
        fitted_sampled_network = self.model_[0]

        genotype = Network.genotype(fitted_sampled_network).normal
        (
            _,
            _,
            param_list,
        ) = fitted_sampled_network.count_parameters()

        if input_labels is not None:
            input_labels_ = tuple(input_labels)
        else:
            input_labels_ = self._get_input_labels()

        assert self.y_ is not None
        out_dim = 1 if self.y_.ndim == 1 else self.y_.shape[1]

        out_func = get_output_str(ValueType(self.output_type))

        # call to plot function
        graph = darts_model_plot(
            genotype=genotype,
            input_labels=input_labels_,
            param_list=param_list,
            full_label=True,
            out_dim=out_dim,
            out_fnc=out_func,
        )

        return graph

    def _get_input_labels(self):
        """
        Returns the input labels for the model.

        Returns:
            input_labels: labels for the input nodes

        """
        return self._get_labels(self.X_, "x")

    def _get_output_labels(self):
        """
        Returns the output labels for the model.

        Returns:
            output_labels: labels for the output nodes

        """
        return self._get_labels(self.y_, "y")

    def _get_labels(
        self, data: Optional[np.ndarray], default_label: str
    ) -> Sequence[str]:
        """
        Returns the labels for the model.

        Arguments:
            data: data to get labels for
            default_label: default label to use if no labels are provided

        Returns:
            labels: labels for the model

        """
        assert data is not None

        if hasattr(data, "columns"):  # it's a dataframe with column names
            labels_ = tuple(data.columns)
        elif (
            hasattr(data, "name") and len(data.shape) == 1
        ):  # it's a single series with a single name
            labels_ = (data.name,)

        else:
            dim = 1 if data.ndim == 1 else data.shape[1]
            labels_ = tuple(f"{default_label}{i+1}" for i in range(dim))
        return labels_

    def model_repr(
        self,
        input_labels: Optional[Sequence[str]] = None,
        output_labels: Optional[Sequence[str]] = None,
        output_function_label: str = "",
        decimals_to_display: int = 2,
        output_format: Literal["latex", "console"] = "console",
    ) -> str:
        """
        Prints the equations of the model architecture.

        Args:
            input_labels: which names to use for the independent variables (X)
            output_labels: which names to use for the dependent variables (y)
            output_function_label: name to use for the output transformation
            decimals_to_display: amount of rounding for the coefficient values
            output_format: whether the output should be formatted for
                the command line (`console`) or as equations in a latex file (`latex`)

        Returns:
            The equations of the model architecture

        """
        assert self.model_ is not None
        fitted_sampled_network: Network = self.model_[0]

        if input_labels is None:
            input_labels_ = self._get_input_labels()
        else:
            input_labels_ = input_labels

        if output_labels is None:
            output_labels_ = self._get_output_labels()
        else:
            output_labels_ = output_labels

        edge_list = fitted_sampled_network.architecture_to_str_list(
            input_labels=input_labels_,
            output_labels=output_labels_,
            output_function_label=output_function_label,
            decimals_to_display=decimals_to_display,
            output_format=output_format,
        )

        model_repr_ = "\n".join(["Model:"] + edge_list)
        return model_repr_


class DARTSExecutionMonitor:
    """
    A monitor of the execution of the DARTS algorithm.
    """

    def __init__(self):
        """
        Initializes the execution monitor.
        """
        self.arch_weight_history = list()
        self.loss_history = list()
        self.epoch_history = list()
        self.primitives = list()

    def execution_monitor(
        self,
        network: Network,
        architect: Architect,
        epoch: int,
        **kwargs: Any,
    ):
        """
        A function to monitor the execution of the DARTS algorithm.

        Arguments:
            network: The DARTS network containing the weights each operation
                in the mixture architecture
            architect: The architect object used to construct the mixture architecture.
            epoch: The current epoch of the training.
            **kwargs: other parameters which may be passed from the DARTS optimizer
        """

        # collect data for visualization
        self.epoch_history.append(epoch)
        self.arch_weight_history.append(
            network.arch_parameters()[0].detach().numpy().copy()[np.newaxis, :]
        )
        self.loss_history.append(architect.current_loss)
        self.primitives = network.primitives

    def display(self):
        """
        A function to display the execution monitor. This function will generate two plots:
        (1) A plot of the training loss vs. epoch,
        (2) a plot of the architecture weights vs. epoch, divided into subplots by each edge
        in the mixture architecture.
        """

        loss_fig, loss_ax = plt.subplots(1, 1)
        loss_ax.plot(self.loss_history)

        loss_ax.set_ylabel("Loss", fontsize=14)
        loss_ax.set_xlabel("Epoch", fontsize=14)
        loss_ax.set_title("Training Loss")

        arch_weight_history_array = np.vstack(self.arch_weight_history)
        num_epochs, num_edges, num_primitives = arch_weight_history_array.shape

        subplots_per_side = int(np.ceil(np.sqrt(num_edges)))

        arch_fig, arch_axes = plt.subplots(
            subplots_per_side,
            subplots_per_side,
            sharex=True,
            sharey=True,
            figsize=(10, 10),
            squeeze=False,
        )

        arch_fig.suptitle("Architecture Weights", fontsize=10)

        for (edge_i, ax) in zip(range(num_edges), arch_axes.flat):
            for primitive_i in range(num_primitives):
                print(f"{edge_i}, {primitive_i}, {ax}")
                ax.plot(
                    arch_weight_history_array[:, edge_i, primitive_i],
                    label=f"{self.primitives[primitive_i]}",
                )

            ax.set_title("k{}".format(edge_i), fontsize=8)

            # there is no need to have the legend for each subplot
            if edge_i == 0:
                ax.legend(loc="upper center")
                ax.set_ylabel("Edge Weights", fontsize=8)
                ax.set_xlabel("Epoch", fontsize=8)

        return SimpleNamespace(
            loss_fig=loss_fig,
            loss_ax=loss_ax,
            arch_fig=arch_fig,
            arch_axes=arch_axes,
        )



import os
import typing

import imageio
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas
import seaborn as sns
import torch.nn
from matplotlib import pyplot
from matplotlib.gridspec import GridSpec

import autora.config as aer_config
import autora.theorist.darts.darts_config as darts_config
from autora.theorist.object_of_study import Object_Of_Study


def generate_darts_summary_figures(
    figure_names: typing.List[str],
    titles: typing.List[str],
    filters: typing.List[str],
    title_suffix: str,
    study_name: str,
    y_name: str,
    y_label: str,
    y_sem_name: str,
    x1_name: str,
    x1_label: str,
    x2_name: str,
    x2_label: str,
    x_limit: typing.List[float],
    y_limit: typing.List[float],
    best_model_name: str,
    figure_size: typing.Tuple[int, int],
    y_reference: typing.List[float] = None,
    y_reference_label: str = "",
    arch_samp_filter: str = None,
):
    """
    Generates a summary figure for a given DARTS study.
    The figure can be composed of different summary plots.

    Arguments:
        figure_names: list of strings with the names of the figures to be generated
        titles: list of strings with the titles of the figures to be generated
        filters: list of strings with the theorist filters to be used to select the models to be
            used in the figures
        title_suffix: string with the suffix to be added to the titles of the figures
        study_name: string with the name of the study (used to identify the study folder)
        y_name: string with the name of the y-axis variable
        y_label: string with the label of the y-axis variable
        y_sem_name: string with the name of the y-axis coding the standard error of the mean
        x1_name: string with the name of the (first) x-axis variable
        x1_label: string with the label of the (first) x-axis variable
        x2_name: string with the name of the second x-axis variable
        x2_label: string with the label of the second x-axis variable
        x_limit: list with the limits of the x-axis
        y_limit: list with the limits of the y-axis
        best_model_name: string with the name of the best model to be highlighted in the figure
        figure_size: list with the size of the figure
        y_reference: list with the values of the reference line
        y_reference_label: string with the label of the reference line
        arch_samp_filter: string with the name of the filter to be used to select the
            samples of the architecture

    """

    for idx, (figure_name, title, theorist_filter) in enumerate(
        zip(figure_names, titles, filters)
    ):

        print("##########################: " + figure_name)
        title = title + title_suffix
        if idx > 0:  # after legend
            show_legend = False
            figure_dimensions = figure_size
        else:
            show_legend = True
            figure_dimensions = (6, 6)
        if idx > 1:  # after original darts
            y_label = " "

        plot_darts_summary(
            study_name=study_name,
            title=title,
            y_name=y_name,
            y_label=y_label,
            y_sem_name=y_sem_name,
            x1_name=x1_name,
            x1_label=x1_label,
            x2_name=x2_name,
            x2_label=x2_label,
            metric="mean_min",
            x_limit=x_limit,
            y_limit=y_limit,
            best_model_name=best_model_name,
            theorist_filter=theorist_filter,
            arch_samp_filter=arch_samp_filter,
            figure_name=figure_name,
            figure_dimensions=figure_dimensions,
            legend_loc=aer_config.legend_loc,
            legend_font_size=aer_config.legend_font_size,
            axis_font_size=aer_config.axis_font_size,
            title_font_size=aer_config.title_font_size,
            show_legend=show_legend,
            y_reference=y_reference,
            y_reference_label=y_reference_label,
            save=True,
        )


def plot_darts_summary(
    study_name: str,
    y_name: str,
    x1_name: str,
    x2_name: str = "",
    y_label: str = "",
    x1_label: str = "",
    x2_label: str = "",
    y_sem_name: str = None,
    metric: str = "min",
    y_reference: typing.List[float] = None,
    y_reference_label: str = "",
    figure_dimensions: typing.Tuple[int, int] = None,
    title: str = "",
    legend_loc: int = 0,
    legend_font_size: int = 8,
    axis_font_size: int = 10,
    title_font_size: int = 10,
    show_legend: bool = True,
    y_limit: typing.List[float] = None,
    x_limit: typing.List[float] = None,
    theorist_filter: str = None,
    arch_samp_filter: str = None,
    best_model_name: str = None,
    save: bool = False,
    figure_name: str = "figure",
):
    """
    Generates a single summary plot for a given DARTS study.

    Arguments:
        study_name: string with the name of the study (used to identify the study folder)
        y_name: string with the name of the y-axis variable
        x1_name: string with the name of the (first) x-axis variable
        x2_name: string with the name of the second x-axis variable
        y_label: string with the label of the y-axis variable
        x1_label: string with the label of the (first) x-axis variable
        x2_label: string with the label of the second x-axis variable
        y_sem_name: string with the name of the y-axis coding the standard error of the mean
        metric: string with the metric to be used to select the best model
        y_reference: list with the values of the reference line
        y_reference_label: string with the label of the reference line
        figure_dimensions: list with the size of the figure
        title: string with the title of the figure
        legend_loc: integer with the location of the legend
        legend_font_size: integer with the font size of the legend
        axis_font_size: integer with the font size of the axis
        title_font_size: integer with the font size of the title
        show_legend: boolean with the flag to show the legend
        y_limit: list with the limits of the y-axis
        x_limit: list with the limits of the x-axis
        theorist_filter: string with the name of the filter to be used to select the theorist
        arch_samp_filter: string with the name of the filter to be used to select the architecture
        best_model_name: string with the name of the best model to be highlighted in the figure
        save: boolean with the flag to save the figure
        figure_name: string with the name of the figure
    """

    palette = "PuBu"

    if figure_dimensions is None:
        figure_dimensions = (4, 3)

    if y_label == "":
        y_label = y_name

    if x1_label == "":
        x1_label = x1_name

    if x2_label == "":
        x2_label = x2_name

    if y_reference_label == "":
        y_reference_label = "Data Generating Model"

    # determine directory for study results and figures
    results_path = (
        aer_config.studies_folder
        + study_name
        + "/"
        + aer_config.models_folder
        + aer_config.models_results_folder
    )

    figures_path = (
        aer_config.studies_folder
        + study_name
        + "/"
        + aer_config.models_folder
        + aer_config.models_results_figures_folder
    )

    # read in all csv files
    files = list()
    for file in os.listdir(results_path):
        if file.endswith(".csv"):
            if "model_" not in file:
                continue

            if theorist_filter is not None:
                if theorist_filter not in file:
                    continue
            files.append(os.path.join(results_path, file))

    print("Found " + str(len(files)) + " files.")

    # generate a plot dictionary
    plot_dict: typing.Dict[typing.Optional[str], typing.List] = dict()
    plot_dict[darts_config.csv_arch_file_name] = list()
    plot_dict[y_name] = list()
    plot_dict[x1_name] = list()
    if x2_name != "":
        plot_dict[x2_name] = list()
    if y_sem_name is not None:
        plot_dict[y_sem_name] = list()

    # load csv files into a common dictionary
    for file in files:
        data = pandas.read_csv(file, header=0)

        valid_data = list()

        # filter for arch samp
        if arch_samp_filter is not None:
            for idx, arch_file_name in enumerate(data[darts_config.csv_arch_file_name]):
                arch_samp = int(
                    float(arch_file_name.split("_sample", 1)[1].split("_", 1)[0])
                )
                if arch_samp == arch_samp_filter:
                    valid_data.append(idx)
        else:
            for idx in range(len(data[darts_config.csv_arch_file_name])):
                valid_data.append(idx)

        plot_dict[darts_config.csv_arch_file_name].extend(
            data[darts_config.csv_arch_file_name][valid_data]
        )
        if y_name in data.keys():
            plot_dict[y_name].extend(data[y_name][valid_data])
        else:
            raise Exception(
                'Could not find key "' + y_name + '" in the data file: ' + str(file)
            )
        if x1_name in data.keys():
            plot_dict[x1_name].extend(data[x1_name][valid_data])
        else:
            raise Exception(
                'Could not find key "' + x1_name + '" in the data file: ' + str(file)
            )
        if x2_name != "":
            if x2_name in data.keys():
                plot_dict[x2_name].extend(data[x2_name][valid_data])
            else:
                raise Exception(
                    'Could not find key "'
                    + x2_name
                    + '" in the data file: '
                    + str(file)
                )
        if y_sem_name is not None:
            # extract seed number from model file name

            if y_sem_name in data.keys():
                plot_dict[y_sem_name].extend(data[y_sem_name])
            elif y_sem_name == "seed":
                y_sem_list = list()
                for file_name in data[darts_config.csv_arch_file_name][valid_data]:
                    y_sem_list.append(
                        int(float(file_name.split("_s_", 1)[1].split("_sample", 1)[0]))
                    )
                plot_dict[y_sem_name].extend(y_sem_list)

            else:

                raise Exception(
                    'Could not find key "'
                    + y_sem_name
                    + '" in the data file: '
                    + str(file)
                )

    model_name_list = plot_dict[darts_config.csv_arch_file_name]
    x1_data = np.asarray(plot_dict[x1_name])
    y_data = np.asarray(plot_dict[y_name])
    if x2_name == "":  # determine for each value of x1 the corresponding y
        x1_data = np.asarray(plot_dict[x1_name])
        x1_unique = np.sort(np.unique(x1_data))

        y_plot = np.empty(x1_unique.shape)
        y_plot[:] = np.nan
        y_sem_plot = np.empty(x1_unique.shape)
        y_sem_plot[:] = np.nan
        y2_plot = np.empty(x1_unique.shape)
        y2_plot[:] = np.nan
        x1_plot = np.empty(x1_unique.shape)
        x1_plot[:] = np.nan
        for idx_unique, x1_unique_val in enumerate(x1_unique):
            y_match = list()
            model_name_match = list()
            for idx_data, x_data_val in enumerate(x1_data):
                if x1_unique_val == x_data_val:
                    y_match.append(y_data[idx_data])
                    model_name_match.append(model_name_list[idx_data])
            x1_plot[idx_unique] = x1_unique_val

            if metric == "min":
                y_plot[idx_unique] = np.min(y_match)
                idx_target = np.argmin(y_match)
                legend_label_spec = " (min)"
            elif metric == "max":
                y_plot[idx_unique] = np.max(y_match)
                idx_target = np.argmax(y_match)
                legend_label_spec = " (max)"
            elif metric == "mean":
                y_plot[idx_unique] = np.mean(y_match)
                idx_target = 0
                legend_label_spec = " (avg)"
            elif metric == "mean_min":
                y_plot[idx_unique] = np.mean(y_match)
                y2_plot[idx_unique] = np.min(y_match)
                idx_target = np.argmin(y_match)
                legend_label_spec = " (avg)"
                legend_label2_spec = " (min)"
            elif metric == "mean_max":
                y_plot[idx_unique] = np.mean(y_match)
                y2_plot[idx_unique] = np.max(y_match)
                idx_target = np.argmax(y_match)
                legend_label_spec = " (avg)"
                legend_label2_spec = " (max)"
            else:
                raise Exception(
                    'Argument "metric" may either be "min", "max", "mean", "mean_min" or "min_max".'
                )

            # compute standard error along given dimension
            if y_sem_name is not None:
                y_sem_data = np.asarray(plot_dict[y_sem_name])
                y_sem_unique = np.sort(np.unique(y_sem_data))
                y_sem = np.empty(y_sem_unique.shape)
                # first average y over all other variables
                for idx_y_sem_unique, y_sem_unique_val in enumerate(y_sem_unique):
                    y_sem_match = list()
                    for idx_y_sem, (
                        y_sem_data_val,
                        x1_data_val,
                        y_data_val,
                    ) in enumerate(zip(y_sem_data, x1_data, y_data)):
                        if (
                            y_sem_unique_val == y_sem_data_val
                            and x1_unique_val == x1_data_val
                        ):
                            y_sem_match.append(y_data_val)
                    y_sem[idx_y_sem_unique] = np.mean(y_sem_match)
                # now compute sem
                y_sem_plot[idx_unique] = np.nanstd(y_sem) / np.sqrt(len(y_sem))

            print(
                x1_label
                + " = "
                + str(x1_unique_val)
                + " ("
                + str(y_plot[idx_unique])
                + "): "
                + model_name_match[idx_target]
            )

    else:  # determine for each combination of x1 and x2 (unique rows) the lowest y
        x2_data = np.asarray(plot_dict[x2_name])
        x2_unique = np.sort(np.unique(x2_data))

        y_plot = list()
        y_sem_plot = list()
        y2_plot = list()
        x1_plot = list()
        x2_plot = list()
        for idx_x2_unique, x2_unique_val in enumerate(x2_unique):

            # collect all x1 and y values matching the current x2 value
            model_name_x2_match = list()
            y_x2_match = list()
            x1_x2_match = list()
            for idx_x2_data, x2_data_val in enumerate(x2_data):
                if x2_unique_val == x2_data_val:
                    model_name_x2_match.append(model_name_list[idx_x2_data])
                    y_x2_match.append(y_data[idx_x2_data])
                    x1_x2_match.append(x1_data[idx_x2_data])

            # now determine unique x1 values for current x2 value
            x1_unique = np.sort(np.unique(x1_x2_match))
            x1_x2_plot = np.empty(x1_unique.shape)
            x1_x2_plot[:] = np.nan
            y_x2_plot = np.empty(x1_unique.shape)
            y_x2_plot[:] = np.nan
            y_sem_x2_plot = np.empty(x1_unique.shape)
            y_sem_x2_plot[:] = np.nan
            y2_x2_plot = np.empty(x1_unique.shape)
            y2_x2_plot[:] = np.nan
            for idx_x1_unique, x1_unique_val in enumerate(x1_unique):
                y_x2_x1_match = list()
                model_name_x2_x1_match = list()
                for idx_x1_data, x1_data_val in enumerate(x1_x2_match):
                    if x1_unique_val == x1_data_val:
                        model_name_x2_x1_match.append(model_name_x2_match[idx_x1_data])
                        y_x2_x1_match.append(y_x2_match[idx_x1_data])
                x1_x2_plot[idx_x1_unique] = x1_unique_val

                if metric == "min":
                    y_x2_plot[idx_x1_unique] = np.min(y_x2_x1_match)
                    idx_target = np.argmin(y_x2_x1_match)
                    legend_label_spec = " (min)"
                elif metric == "max":
                    y_x2_plot[idx_x1_unique] = np.max(y_x2_x1_match)
                    idx_target = np.argmax(y_x2_x1_match)
                    legend_label_spec = " (max)"
                elif metric == "mean":
                    y_x2_plot[idx_x1_unique] = np.mean(y_x2_x1_match)
                    idx_target = 0
                    legend_label_spec = " (avg)"
                elif metric == "mean_min":
                    y_x2_plot[idx_x1_unique] = np.mean(y_x2_x1_match)
                    y2_x2_plot[idx_x1_unique] = np.min(y_x2_x1_match)
                    idx_target = np.argmin(y_x2_x1_match)
                    legend_label_spec = " (avg)"
                    legend_label2_spec = " (min)"
                elif metric == "mean_max":
                    y_x2_plot[idx_x1_unique] = np.mean(y_x2_x1_match)
                    y2_x2_plot[idx_x1_unique] = np.max(y_x2_x1_match)
                    idx_target = np.argmax(y_x2_x1_match)
                    legend_label_spec = " (avg)"
                    legend_label2_spec = " (max)"
                else:
                    raise Exception(
                        'Argument "metric" may either be "min", "max", "mean", '
                        '"mean_min" or "min_max".'
                    )

                # compute standard error along given dimension
                if y_sem_name is not None:
                    y_sem_data = np.asarray(plot_dict[y_sem_name])
                    y_sem_unique = np.sort(np.unique(y_sem_data))
                    y_sem = np.empty(y_sem_unique.shape)
                    # first average y over all other variables
                    for idx_y_sem_unique, y_sem_unique_val in enumerate(y_sem_unique):
                        y_sem_match = list()
                        for idx_y_sem, (
                            y_sem_data_val,
                            x1_data_val,
                            x2_data_val,
                            y_data_val,
                        ) in enumerate(zip(y_sem_data, x1_data, x2_data, y_data)):
                            if (
                                y_sem_unique_val == y_sem_data_val
                                and x1_unique_val == x1_data_val
                                and x2_unique_val == x2_data_val
                            ):
                                y_sem_match.append(y_data_val)
                        y_sem[idx_y_sem_unique] = np.nanmean(y_sem_match)
                    # now compute sem
                    y_sem_x2_plot[idx_x1_unique] = np.nanstd(y_sem) / np.sqrt(
                        len(y_sem)
                    )

                if metric == "mean_min" or metric == "mean_max":
                    best_val_str = str(y2_x2_plot[idx_x1_unique])
                else:
                    best_val_str = str(y_x2_plot[idx_x1_unique])

                print(
                    x1_label
                    + " = "
                    + str(x1_unique_val)
                    + ", "
                    + x2_label
                    + " = "
                    + str(x2_unique_val)
                    + " ("
                    + best_val_str
                    + "): "
                    + model_name_x2_x1_match[idx_target]
                )

            y_plot.append(y_x2_plot)
            y2_plot.append(y2_x2_plot)
            y_sem_plot.append(y_sem_x2_plot)
            x1_plot.append(x1_x2_plot)
            x2_plot.append(x2_unique_val)
    # plot
    # plt.axhline

    # determine best model coordinates
    best_model_x1 = None
    best_model_x2 = None
    best_model_y = None
    if best_model_name is not None:
        theorist = best_model_name.split("weights_", 1)[1].split("_v_", 1)[0]
        if theorist_filter is not None:
            if theorist_filter == theorist:
                determine_best_model = True
            else:
                determine_best_model = False
        else:
            determine_best_model = True

        if determine_best_model:
            idx = plot_dict[darts_config.csv_arch_file_name].index(best_model_name)
            best_model_x1 = plot_dict[x1_name][idx]
            best_model_x2 = plot_dict[x2_name][idx]
            best_model_y = plot_dict[y_name][idx]

    fig, ax = pyplot.subplots(figsize=figure_dimensions)

    if x2_name == "":

        colors = sns.color_palette(palette, 10)
        color = colors[-1]
        full_label = "Reconstructed Model" + legend_label_spec
        sns.lineplot(
            x=x1_plot,
            y=y_plot,
            marker="o",
            linewidth=2,
            ax=ax,
            label=full_label,
            color=color,
        )

        # draw error bars
        if y_sem_name is not None:
            ax.errorbar(x=x1_plot, y=y_plot, yerr=y_sem_plot, color=color)

        # draw second y value
        if metric == "mean_min" or metric == "mean_max":
            full_label = "Reconstructed Model" + legend_label2_spec
            ax.plot(x1_plot, y2_plot, "*", linewidth=2, label=full_label, color=color)

            if show_legend:
                handles, _ = ax.get_legend_handles_labels()
                ax.legend(handles=handles, loc=legend_loc)
                plt.setp(ax.get_legend().get_texts(), fontsize=legend_font_size)

        # draw selected model
        if best_model_x1 is not None and best_model_y is not None:
            ax.plot(
                best_model_x1,
                best_model_y,
                "o",
                fillstyle="none",
                color="black",
                markersize=10,
            )

        ax.set_xlabel(x1_label, fontsize=axis_font_size)
        ax.set_ylabel(y_label, fontsize=axis_font_size)
        ax.set_title(title, fontsize=title_font_size)

        if y_limit is not None:
            ax.set_ylim(y_limit)

        if x_limit is not None:
            ax.set_xlim(x_limit)

        # generate legend
        # ax.scatter(x1_plot, y_plot, marker='.', c='r')
        # g = sns.relplot(data=data_plot, x=x1_label, y=y_label, ax=ax)
        # g._legend.remove()
        if y_reference is not None:
            ax.axhline(
                y_reference, c="black", linestyle="dashed", label=y_reference_label
            )

            if show_legend:
                # generate legend
                handles, _ = ax.get_legend_handles_labels()
                ax.legend(handles=handles, loc=legend_loc)
                plt.setp(ax.get_legend().get_texts(), fontsize=legend_font_size)
    else:

        colors = sns.color_palette(palette, len(x2_plot))

        for idx, x2 in enumerate(x2_plot):

            x1_plot_line = x1_plot[idx]
            y_plot_line = y_plot[idx]
            label = x2_label + "$ = " + str(x2) + "$" + legend_label_spec
            color = colors[idx]

            sns.lineplot(
                x=x1_plot_line,
                y=y_plot_line,
                marker="o",
                linewidth=2,
                ax=ax,
                label=label,
                color=color,
                alpha=1,
            )

            # draw error bars
            if y_sem_name is not None:
                y_sem_plot_line = y_sem_plot[idx]
                ax.errorbar(
                    x=x1_plot_line,
                    y=y_plot_line,
                    yerr=y_sem_plot_line,
                    color=color,
                    alpha=1,
                )

        # # draw second y value on top
        # for idx, x2 in enumerate(x2_plot):
        #     x1_plot_line = x1_plot[idx]
        #     color = colors[idx]
        #
        #     if metric == 'mean_min' or metric == 'mean_max':
        #         y2_plot_line = y2_plot[idx]
        #         label = x2_label + '$ = ' + str(x2) + "$" + legend_label2_spec
        #         ax.plot(x1_plot_line, y2_plot_line, '*', linewidth=2, label=label, color=color)

        # draw selected model
        if best_model_x1 is not None and best_model_y is not None:
            ax.plot(
                best_model_x1,
                best_model_y,
                "o",
                fillstyle="none",
                color="black",
                markersize=10,
            )

            for idx, x2 in enumerate(x2_plot):
                if best_model_x2 == x2:
                    color = colors[idx]
            ax.plot(
                best_model_x1,
                best_model_y,
                "*",
                linewidth=2,
                label="Best Model",
                color=color,
            )

        if y_reference is not None:
            ax.axhline(
                y_reference, c="black", linestyle="dashed", label=y_reference_label
            )

        handles, _ = ax.get_legend_handles_labels()
        leg = ax.legend(
            handles=handles, loc=legend_loc, bbox_to_anchor=(1.05, 1)
        )  # , title='Legend'
        plt.setp(ax.get_legend().get_texts(), fontsize=legend_font_size)

        if not show_legend:
            leg.remove()

        if y_limit is not None:
            ax.set_ylim(y_limit)

        if x_limit is not None:
            ax.set_xlim(x_limit)

    sns.despine(trim=True)
    ax.set_ylabel(y_label, fontsize=axis_font_size)
    ax.set_xlabel(x1_label, fontsize=axis_font_size)
    ax.set_title(title, fontsize=title_font_size)
    plt.show()

    # save plot
    if save:
        if not os.path.exists(figures_path):
            os.mkdir(figures_path)
        fig.savefig(os.path.join(figures_path, figure_name))


def plot_model_graph(
    study_name: str,
    arch_weights_name: str,
    model_weights_name: str,
    object_of_study: Object_Of_Study,
    figure_name: str = "graph",
):
    """
    Plot the graph of the DARTS model.

    Arguments:
        study_name: name of the study (used to identify the relevant study folder)
        arch_weights_name: name of the architecture weights file
        model_weights_name: name of the model weights file (that contains the trained parameters)
        object_of_study: name of the object of study
        figure_name: name of the figure
    """

    import os

    import autora.theorist.darts.utils as utils
    import autora.theorist.darts.visualize as viz

    figures_path = (
        aer_config.studies_folder
        + study_name
        + "/"
        + aer_config.models_folder
        + aer_config.models_results_figures_folder
    )

    model = load_model(
        study_name, model_weights_name, arch_weights_name, object_of_study
    )

    (n_params_total, n_params_base, param_list) = model.countParameters(
        print_parameters=True
    )
    genotype = model.genotype()
    filepath = os.path.join(figures_path, figure_name)
    viz.plot(
        genotype.normal,
        filepath,
        file_format="png",
        view_file=True,
        full_label=True,
        param_list=param_list,
        input_labels=object_of_study.__get_input_labels__(),
        out_dim=object_of_study.__get_output_dim__(),
        out_fnc=utils.get_output_str(object_of_study.__get_output_type__()),
    )


# old


def load_model(
    study_name: str,
    model_weights_name: str,
    arch_weights_name: str,
    object_of_study: Object_Of_Study,
) -> torch.nn.Module:
    """
    Load the model.

    Arguments:
        study_name: name of the study (used to identify the relevant study folder)
        model_weights_name: name of the model weights file (that contains the trained parameters)
        arch_weights_name: name of the architecture weights file
        object_of_study: name of the object of study

    Returns:
        model: DARTS model
    """

    import os

    import torch

    import autora.theorist.darts.utils as utils
    from autora.theorist.darts.model_search import Network

    num_output = object_of_study.__get_output_dim__()
    num_input = object_of_study.__get_input_dim__()
    k = int(float(arch_weights_name.split("_k_", 1)[1].split("_s_", 1)[0]))

    results_weights_path = (
        aer_config.studies_folder
        + study_name
        + "/"
        + aer_config.models_folder
        + aer_config.models_results_weights_folder
    )

    model_path = os.path.join(results_weights_path, model_weights_name + ".pt")
    arch_path = os.path.join(results_weights_path, arch_weights_name + ".pt")
    criterion = utils.sigmid_mse
    model = Network(num_output, criterion, steps=k, n_input_states=num_input)
    utils.load(model, model_path)
    alphas_normal = torch.load(arch_path)
    model.fix_architecture(True, new_weights=alphas_normal)

    return model


class DebugWindow:
    """
    A window with plots that are used for debugging.
    """

    def __init__(
        self,
        num_epochs: int,
        numArchEdges: int = 1,
        numArchOps: int = 1,
        ArchOpsLabels: typing.Tuple = (),
        fitPlot3D: bool = False,
        show_arch_weights: bool = True,
    ):
        """
        Initializes the debug window.

        Arguments:
            num_epochs: number of architecture training epochs
            numArchEdges: number of architecture edges
            numArchOps: number of architecture operations
            ArchOpsLabels: list of architecture operation labels
            fitPlot3D: if True, the 3D plot of the fit is shown
            show_arch_weights: if True, the architecture weights are shown
        """

        # initialization
        matplotlib.use("TkAgg")  # need to add this for PyCharm environment

        plt.ion()

        # SETTINGS
        self.show_arch_weights = show_arch_weights
        self.fontSize = 10

        self.performancePlot_limit = (0, 1)
        self.modelFitPlot_limit = (0, 500)
        self.mismatchPlot_limit = (0, 1)
        self.architectureWeightsPlot_limit = (0.1, 0.2)

        self.numPatternsShown = 100

        # FIGURE
        self.fig = plt.figure()
        self.fig.set_size_inches(13, 7)

        if self.show_arch_weights is False:
            numArchEdges = 0

        # set up grid
        numRows = np.max((1 + np.ceil((numArchEdges + 1) / 4), 2))
        gs = GridSpec(numRows.astype(int), 4, figure=self.fig)

        self.fig.subplots_adjust(
            left=0.1, bottom=0.1, right=0.90, top=0.9, wspace=0.4, hspace=0.5
        )
        self.modelGraph = self.fig.add_subplot(gs[1, 0])
        self.performancePlot = self.fig.add_subplot(gs[0, 0])
        self.modelFitPlot = self.fig.add_subplot(gs[0, 1])
        if fitPlot3D:
            self.mismatchPlot = self.fig.add_subplot(gs[0, 2], projection="3d")
        else:
            self.mismatchPlot = self.fig.add_subplot(gs[0, 2])
        self.examplePatternsPlot = self.fig.add_subplot(gs[0, 3])

        self.architecturePlot = []

        for edge in range(numArchEdges):
            row = np.ceil((edge + 2) / 4).astype(int)
            col = (edge + 1) % 4
            self.architecturePlot.append(self.fig.add_subplot(gs[row, col]))

        self.colors = (
            "black",
            "red",
            "green",
            "blue",
            "purple",
            "orange",
            "brown",
            "pink",
            "grey",
            "olive",
            "cyan",
            "yellow",
            "skyblue",
            "coral",
            "magenta",
            "seagreen",
            "sandybrown",
        )

        # PERFORMANCE PLOT
        x = 1
        y = 1
        (self.train_error,) = self.performancePlot.plot(x, y, "r-")
        (self.valid_error,) = self.performancePlot.plot(x, y, "b", linestyle="dashed")

        # set labels
        self.performancePlot.set_xlabel("Epoch", fontsize=self.fontSize)
        self.performancePlot.set_ylabel("Cross-Entropy Loss", fontsize=self.fontSize)
        self.performancePlot.set_title("Performance", fontsize=self.fontSize)
        self.performancePlot.legend(
            (self.train_error, self.valid_error), ("training error", "validation error")
        )

        # adjust axes
        self.performancePlot.set_xlim(0, num_epochs)
        self.performancePlot.set_ylim(
            self.performancePlot_limit[0], self.performancePlot_limit[1]
        )

        # MODEL FIT PLOT
        x = 1
        y = 1
        (self.BIC,) = self.modelFitPlot.plot(x, y, color="black")
        (self.AIC,) = self.modelFitPlot.plot(x, y, color="grey")

        # set labels
        self.modelFitPlot.set_xlabel("Epoch", fontsize=self.fontSize)
        self.modelFitPlot.set_ylabel("Information Criterion", fontsize=self.fontSize)
        self.modelFitPlot.set_title("Model Fit", fontsize=self.fontSize)
        self.modelFitPlot.legend((self.BIC, self.AIC), ("BIC", "AIC"))

        # adjust axes
        self.modelFitPlot.set_xlim(0, num_epochs)
        self.modelFitPlot.set_ylim(
            self.modelFitPlot_limit[0], self.modelFitPlot_limit[1]
        )

        # RANGE PREDICTION FIT PLOT
        x = 1
        y = 1
        if fitPlot3D:
            x = np.arange(0, 1, 0.1)
            y = np.arange(0, 1, 0.1)
            X, Y = np.meshgrid(x, y)
            Z = X * np.exp(-X - Y)

            self.range_target = self.mismatchPlot.plot_surface(X, Y, Z)
            self.range_prediction = self.mismatchPlot.plot_surface(X, Y, Z)
            self.mismatchPlot.set_zlim(
                self.mismatchPlot_limit[0], self.mismatchPlot_limit[1]
            )

            # set labels
            self.mismatchPlot.set_xlabel("Stimulus 1", fontsize=self.fontSize)
            self.mismatchPlot.set_ylabel("Stimulus 2", fontsize=self.fontSize)
            self.mismatchPlot.set_zlabel("Outcome Value", fontsize=self.fontSize)

        else:
            (self.range_target,) = self.mismatchPlot.plot(x, y, color="black")
            (self.range_prediction,) = self.mismatchPlot.plot(x, y, "--", color="red")

            # set labels
            self.mismatchPlot.set_xlabel("Stimulus Value", fontsize=self.fontSize)
            self.mismatchPlot.set_ylabel("Outcome Value", fontsize=self.fontSize)
            self.mismatchPlot.legend(
                (self.range_target, self.range_prediction), ("target", "prediction")
            )

        self.mismatchPlot.set_title("Target vs. Prediction", fontsize=self.fontSize)

        # adjust axes
        self.mismatchPlot.set_xlim(0, 1)
        self.mismatchPlot.set_ylim(0, 1)

        # ARCHITECTURE WEIGHT PLOT
        if self.show_arch_weights:

            self.architectureWeights = []
            for idx, architecturePlot in enumerate(self.architecturePlot):
                plotWeights = []
                x = 1
                y = 1
                for op in range(numArchOps):
                    (plotWeight,) = architecturePlot.plot(x, y, color=self.colors[op])
                    plotWeights.append(plotWeight)

                # set legend
                if idx == 0:
                    architecturePlot.legend(
                        plotWeights, ArchOpsLabels, prop={"size": 6}
                    )

                # add labels
                architecturePlot.set_ylabel("Weight", fontsize=self.fontSize)
                architecturePlot.set_title(
                    "(" + str(idx) + ") Edge Weight", fontsize=self.fontSize
                )
                if idx == len(self.architecturePlot) - 1:
                    architecturePlot.set_xlabel("Epoch", fontsize=self.fontSize)

                # adjust axes
                architecturePlot.set_xlim(0, num_epochs)
                architecturePlot.set_ylim(
                    self.architectureWeightsPlot_limit[0],
                    self.architectureWeightsPlot_limit[1],
                )

                self.architectureWeights.append(plotWeights)

        # draw
        plt.draw()

    def update(
        self,
        train_error: np.array = None,
        valid_error: np.array = None,
        weights: np.array = None,
        BIC: np.array = None,
        AIC: np.array = None,
        model_graph: str = None,
        range_input1: np.array = None,
        range_input2: np.array = None,
        range_target: np.array = None,
        range_prediction: np.array = None,
        target: np.array = None,
        prediction: np.array = None,
    ):
        """
        Update the debug plot with new data.

        Arguments:
            train_error: training error
            valid_error: validation error
            weights: weights of the model
            BIC: Bayesian information criterion of the model
            AIC: Akaike information criterion of the model
            model_graph: the graph of the model
            range_input1: the range of the first input
            range_input2: the range of the second input
            range_target: the range of the target
            range_prediction: the range of the prediction
            target: the target
            prediction: the prediction
        """

        # update training error
        if train_error is not None:
            self.train_error.set_xdata(
                np.linspace(1, len(train_error), len(train_error))
            )
            self.train_error.set_ydata(train_error)

        # update validation error
        if valid_error is not None:
            self.valid_error.set_xdata(
                np.linspace(1, len(valid_error), len(valid_error))
            )
            self.valid_error.set_ydata(valid_error)

        # update BIC
        if BIC is not None:
            self.BIC.set_xdata(np.linspace(1, len(BIC), len(BIC)))
            self.BIC.set_ydata(BIC)

        # update AIC
        if AIC is not None:
            self.AIC.set_xdata(np.linspace(1, len(AIC), len(AIC)))
            self.AIC.set_ydata(AIC)

        # update target vs. prediction plot
        if (
            range_input1 is not None
            and range_target is not None
            and range_prediction is not None
            and range_input2 is None
        ):
            self.range_target.set_xdata(range_input1)
            self.range_target.set_ydata(range_target)
            self.range_prediction.set_xdata(range_input1)
            self.range_prediction.set_ydata(range_prediction)
        elif (
            range_input1 is not None
            and range_target is not None
            and range_prediction is not None
            and range_input2 is not None
        ):

            # update plot
            self.mismatchPlot.cla()
            self.range_target = self.mismatchPlot.plot_surface(
                range_input1, range_input2, range_target, color=(0, 0, 0, 0.5)
            )
            self.range_prediction = self.mismatchPlot.plot_surface(
                range_input1, range_input2, range_prediction, color=(1, 0, 0, 0.5)
            )

            # set labels
            self.mismatchPlot.set_xlabel("Stimulus 1", fontsize=self.fontSize)
            self.mismatchPlot.set_ylabel("Stimulus 2", fontsize=self.fontSize)
            self.mismatchPlot.set_zlabel("Outcome Value", fontsize=self.fontSize)
            self.mismatchPlot.set_title("Target vs. Prediction", fontsize=self.fontSize)

        # update example pattern plot
        if target is not None and prediction is not None:

            # select limited number of patterns
            self.numPatternsShown = np.min((self.numPatternsShown, target.shape[0]))
            target = target[0 : self.numPatternsShown, :]
            prediction = prediction[0 : self.numPatternsShown, :]

            im = np.concatenate((target, prediction), axis=1)
            self.examplePatternsPlot.cla()
            self.examplePatternsPlot.imshow(im, interpolation="nearest", aspect="auto")
            x = np.ones(target.shape[0]) * (target.shape[1] - 0.5)
            y = np.linspace(1, target.shape[0], target.shape[0])
            self.examplePatternsPlot.plot(x, y, color="red")

            # set labels
            self.examplePatternsPlot.set_xlabel("Output", fontsize=self.fontSize)
            self.examplePatternsPlot.set_ylabel("Pattern", fontsize=self.fontSize)
            self.examplePatternsPlot.set_title(
                "Target vs. Prediction", fontsize=self.fontSize
            )

        if self.show_arch_weights:
            # update weights
            if weights is not None:
                for plotIdx, architectureWeights in enumerate(self.architectureWeights):
                    for lineIdx, plotWeight in enumerate(architectureWeights):
                        plotWeight.set_xdata(
                            np.linspace(1, weights.shape[0], weights.shape[0])
                        )
                        plotWeight.set_ydata(weights[:, plotIdx, lineIdx])

        # draw current graph
        if model_graph is not None:
            im = imageio.imread(model_graph)
            self.modelGraph.cla()
            self.modelGraph.imshow(im)
            self.modelGraph.axis("off")

        # re-draw plot
        plt.draw()
        plt.pause(0.02)



from abc import abstractmethod
from typing import Sequence

from tinkerforge.bricklet_industrial_analog_out_v2 import BrickletIndustrialAnalogOutV2
from tinkerforge.bricklet_industrial_dual_0_20ma_v2 import BrickletIndustrialDual020mAV2
from tinkerforge.bricklet_industrial_dual_analog_in_v2 import (
    BrickletIndustrialDualAnalogInV2,
)
from tinkerforge.ip_connection import IPConnection
from variable import ValueType

from autora.variable import DV, IV, Variable, register_dv_label, register_iv_label


class TinkerforgeVariable(Variable):
    """
    A representation of a variable used in the Tinkerforge environment.
    """

    _variable_label = ""
    _UID = ""
    _priority = 0

    def __init__(
        self,
        variable_label: str = "",
        UID: str = "",
        name: str = "",
        units: str = "",
        priority: int = 0,
        value_range: Sequence = (0, 1),
        type: ValueType = float,
    ):
        """
        Initializes a Tinkerforge variable.
        Args:
            variable_label: the label of the variable
            UID: the user identification of the variable
            name: the name of the variable
            units: the units of the variable
            priority: the priority of the variable
            value_range: the value range of the variable
            type: the type of the variable
        """

        super().__init__(
            name=name,
            value_range=value_range,
            units=units,
            type=type,
            variable_label=variable_label,
        )

        self._UID = UID
        self._priority = priority

    def __get_priority__(self) -> int:
        """
        Get priority of variable. The priority is used to determine the sequence of variables
        to be measured or manipulated.

        Returns:
            The priority of the variable.
        """
        return self._priority

    def __set_priority__(self, priority: int = 0):
        """
        Set priority of variable.
        The priority is used to determine the sequence of variables to be measured or manipulated.

        Arguments:
            priority: The priority of the variable.
        """
        self._priority = priority

    @abstractmethod
    def clean_up(self):
        """Clean up measurement device."""
        pass

    @abstractmethod
    def disconnect(self):
        """Disconnect from up measurement device."""
        pass


class IVTF(IV, TinkerforgeVariable):
    """
    A representation of an independent variable used in the Tinkerforge environment.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes an independent variable used in the Tinkerforge environment.

        For arguments, see [autora.variable.tinkerforge.TinkerforgeVariable]
        [autora.variable.tinkerforge.TinkerforgeVariable.__init__]
        """
        IV.__init__(self, *args, **kwargs)
        TinkerforgeVariable.__init__(self, *args, **kwargs)


class DVTF(DV, TinkerforgeVariable):
    """
    A representation of a dependent variable used in the Tinkerforge environment.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes a dependent variable used in the Tinkerforge environment.

        For arguments, see [autora.variable.tinkerforge.TinkerforgeVariable]
        [autora.variable.tinkerforge.TinkerforgeVariable.__init__]
        """
        DV.__init__(self, *args, **kwargs)
        TinkerforgeVariable.__init__(self, *args, **kwargs)


class IVCurrent(IVTF):
    """
    An independent tinkerforge variable representing the current.
    """

    _name = "source_current"
    _UID = "MST"
    _variable_label = "Source Current"
    _units = "µA"
    _priority = 0
    _value_range = (0, 20000)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223

    def __init__(self, *args, **kwargs):
        """
        Initializes Industrial Analog Out 2.0 device.

        For arguments, see [autora.variable.tinkerforge.TinkerforgeVariable]
        [autora.variable.tinkerforge.TinkerforgeVariable.__init__]
        """

        self._ipcon = IPConnection()  # Create IP connection
        self._iao = BrickletIndustrialAnalogOutV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        super(IVCurrent, self).__init__(*args, **kwargs)

    def disconnect(self):
        """
        Disconnect from up measurement device.
        """

        self._iao.set_enabled(False)

        self._ipcon.disconnect()

    def stop(self):
        """
        Disable current output
        """

        self._iao.set_enabled(False)

    def manipulate(self):
        """
        Sets the current output to the specified value.
        """
        self._iao.set_current(self.get_value())
        self._iao.set_enabled(True)

    def clean_up(self):
        """
        Clean up measurement device.
        """
        self.stop()


class IVVoltage(IVTF):
    """
    An independent tinkerforge variable representing the voltage.
    """

    _variable_label = "Source Voltage"
    _UID = "MST"
    _name = "source_voltage"
    _units = "mV"
    _priority = 0
    _value_range = (0, 5000)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223

    def __init__(self, *args, **kwargs):
        """
        Initializes Industrial Analog Out 2.0 device.
        """

        self._ipcon = IPConnection()  # Create IP connection
        self._iao = BrickletIndustrialAnalogOutV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        super(IVVoltage, self).__init__(*args, **kwargs)

    def disconnect(self):
        """
        Disconnect from up measurement device.
        """

        self._iao.set_enabled(False)

        self._ipcon.disconnect()

    def stop(self):
        """
        Disable voltage output
        """
        self._iao.set_enabled(False)

    def manipulate(self):
        """
        Sets the voltage output to the specified value.
        """
        self._iao.set_voltage(self.get_value())
        self._iao.set_enabled(True)

    def clean_up(self):
        """
        Clean up measurement device.
        """
        self.stop()


class DVCurrent(DVTF):
    """
    A dependent tinkerforge variable representing the current.
    """

    _name = "current0"
    _UID = "Hfg"
    _variable_label = "Current 0"
    _units = "mA"
    _priority = 0
    _value_range = (0, 2000)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223
    channel = 0

    def __init__(self, *args, **kwargs):
        """
        Initializes Industrial Analog Out 2.0 device.

        For arguments, see [autora.variable.tinkerforge.TinkerforgeVariable]
        [autora.variable.tinkerforge.TinkerforgeVariable.__init__]
        """

        super(DVCurrent, self).__init__(*args, **kwargs)

        self._ipcon = IPConnection()  # Create IP connection
        self._id020 = BrickletIndustrialDual020mAV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        if self._name == "current1":
            self.channel = 1
        else:
            self.channel = 0

    def disconnect(self):
        """
        Disconnect from up measurement device.
        """

        self._ipcon.disconnect()

    def measure(self):
        """
        Measures the current.
        """
        current = self._id020.get_current(self.channel)
        self.set_value(current / 1000000.0)


class DVVoltage(DVTF):
    """
    A dependent tinkerforge variable representing the voltage.
    """

    _name = "voltage0"
    _UID = "MjY"
    _variable_label = "Voltage 0"
    _units = "mV"
    _priority = 0
    _value_range = (-3500, 3500)
    _value = 0

    _HOST = "localhost"
    _PORT = 4223

    channel = 0

    def __init__(self, *args, **kwargs):
        """
        Initializes Industrial Analog Out 2.0 device.

        For arguments, see [autora.variable.tinkerforge.TinkerforgeVariable]
        [autora.variable.tinkerforge.TinkerforgeVariable.__init__]
        """

        super(DVVoltage, self).__init__(*args, **kwargs)

        self._ipcon = IPConnection()  # Create IP connection
        self._idai = BrickletIndustrialDualAnalogInV2(
            self._UID, self._ipcon
        )  # Create device object

        self._ipcon.connect(self._HOST, self._PORT)  # Connect to brickd

        if self._name == "voltage1":
            self.channel = 1
        else:
            self.channel = 0

    def disconnect(self):
        """
        Disconnect from up measurement device.
        """
        self._ipcon.disconnect()

    def measure(self):
        """
        Measures the voltage.
        """
        value = self._idai.get_voltage(self.channel)
        self.set_value(value)


register_dv_label(
    **{
        "voltage0": (DVVoltage, "Voltage 0", "MjY", "voltage0", "mV", 1, (-3500, 3500)),
        "voltage1": (DVVoltage, "Voltage 1", "MjY", "voltage1", "mV", 1, (-3500, 3500)),
        "current0": (DVCurrent, "Current 0", "Hfg", "current0", "mA", 2, (0, 20)),
        "current1": (DVCurrent, "Current 1", "Hfg", "current1", "mA", 2, (0, 20)),
    }
)
register_iv_label(
    **{
        "source_voltage": (
            IVVoltage,
            "Source Voltage",
            "MST",
            "source_voltage",
            "mV",
            2,
            (0, 5000),
        ),
        "source_current": (
            IVCurrent,
            "Source Current",
            "MST",
            "source_current",
            "µA",
            2,
            (0, 20000),
        ),
    }
)



import typing
from collections import namedtuple

import torch
import torch.nn as nn

Genotype = namedtuple("Genotype", "normal normal_concat")


def isiterable(p_object: typing.Any) -> bool:
    """
    Checks if an object is iterable.

    Arguments:
        p_object: object to be checked
    """
    try:
        iter(p_object)
    except TypeError:
        return False
    return True


def get_operation_label(
    op_name: str,
    params_org: typing.List,
    decimals: int = 4,
    input_var: str = "x",
    output_format: typing.Literal["latex", "console"] = "console",
) -> str:
    r"""
    Returns a complete string describing a DARTS operation.

    Arguments:
        op_name: name of the operation
        params_org: original parameters of the operation
        decimals: number of decimals to be used for converting the parameters into string format
        input_var: name of the input variable
        output_format: format of the output string (either "latex" or "console")

    Examples:
        >>> get_operation_label("classifier", [1], decimals=2)
        '1.00 * x'
        >>> import numpy as np
        >>> print(get_operation_label("classifier_concat", np.array([1, 2, 3]),
        ...     decimals=2, output_format="latex"))
        x \circ \left(1.00\right) + \left(2.00\right) + \left(3.00\right)
        >>> get_operation_label("classifier_concat", np.array([1, 2, 3]),
        ...     decimals=2, output_format="console")
        'x .* (1.00) .+ (2.00) .+ (3.00)'
        >>> get_operation_label("linear_exp", [1,2], decimals=2)
        'exp(1.00 * x + 2.00)'
        >>> get_operation_label("none", [])
        ''
        >>> get_operation_label("reciprocal", [1], decimals=0)
        '1 / x'
        >>> get_operation_label("linear_reciprocal", [1, 2], decimals=0)
        '1 / (1 * x + 2)'
        >>> get_operation_label("linear_relu", [1], decimals=0)
        'ReLU(1 * x)'
        >>> print(get_operation_label("linear_relu", [1], decimals=0, output_format="latex"))
        \operatorname{ReLU}\left(1x\right)
        >>> get_operation_label("linear", [1, 2], decimals=0)
        '1 * x + 2'
        >>> get_operation_label("linear", [1, 2], decimals=0, output_format="latex")
        '1 x + 2'
        >>> get_operation_label("linrelu", [1], decimals=0)  # Mistyped operation name
        Traceback (most recent call last):
        ...
        NotImplementedError: operation 'linrelu' is not defined for output_format 'console'
    """
    if output_format != "latex" and output_format != "console":
        raise ValueError("output_format must be either 'latex' or 'console'")

    params = params_org.copy()

    format_string = "{:." + "{:.0f}".format(decimals) + "f}"

    classifier_str = ""
    if op_name == "classifier":
        value = params[0]
        classifier_str = f"{format_string.format(value)} * {input_var}"
        return classifier_str

    if op_name == "classifier_concat":
        if output_format == "latex":
            classifier_str = input_var + " \\circ \\left("
        else:
            classifier_str = input_var + " .* ("
        for param_idx, param in enumerate(params):

            if param_idx > 0:
                if output_format == "latex":
                    classifier_str += " + \\left("
                else:
                    classifier_str += " .+ ("

            if isiterable(param.tolist()):

                param_formatted = list()
                for value in param.tolist():
                    param_formatted.append(format_string.format(value))

                for value_idx, value in enumerate(param_formatted):
                    if value_idx < len(param) - 1:
                        classifier_str += value + " + "
                    else:
                        if output_format == "latex":
                            classifier_str += value + "\\right)"
                        else:
                            classifier_str += value + ")"

            else:
                value = format_string.format(param)

                if output_format == "latex":
                    classifier_str += value + "\\right)"
                else:
                    classifier_str += value + ")"

        return classifier_str

    num_params = len(params)

    c = [str(format_string.format(p)) for p in params_org]
    c.extend(["", "", ""])

    if num_params == 1:  # without bias
        if output_format == "console":
            labels = {
                "none": "",
                "add": f"+ {input_var}",
                "subtract": f"- {input_var}",
                "mult": f"{c[0]} * {input_var}",
                "linear": f"{c[0]} * {input_var}",
                "relu": f"ReLU({input_var})",
                "linear_relu": f"ReLU({c[0]} * {input_var})",
                "logistic": f"logistic({input_var})",
                "linear_logistic": f"logistic({c[0]} * {input_var})",
                "exp": f"exp({input_var})",
                "linear_exp": f"exp({c[0]} * {input_var})",
                "reciprocal": f"1 / {input_var}",
                "linear_reciprocal": f"1 / ({c[0]} * {input_var})",
                "ln": f"ln({input_var})",
                "linear_ln": f"ln({c[0]} * {input_var})",
                "cos": f"cos({input_var})",
                "linear_cos": f"cos({c[0]} * {input_var})",
                "sin": f"sin({input_var})",
                "linear_sin": f"sin({c[0]} * {input_var})",
                "tanh": f"tanh({input_var})",
                "linear_tanh": f"tanh({c[0]} * {input_var})",
                "classifier": classifier_str,
            }
        elif output_format == "latex":
            labels = {
                "none": "",
                "add": f"+ {input_var}",
                "subtract": f"- {input_var}",
                "mult": f"{c[0]} {input_var}",
                "linear": c[0] + "" + input_var,
                "relu": f"\\operatorname{{ReLU}}\\left({input_var}\\right)",
                "linear_relu": f"\\operatorname{{ReLU}}\\left({c[0]}{input_var}\\right)",
                "logistic": f"\\sigma\\left({input_var}\\right)",
                "linear_logistic": f"\\sigma\\left({c[0]} {input_var} \\right)",
                "exp": f"+ e^{input_var}",
                "linear_exp": f"e^{{{c[0]} {input_var} }}",
                "reciprocal": f"\\frac{{1}}{{{input_var}}}",
                "linear_reciprocal": f"\\frac{{1}}{{{c[0]} {input_var} }}",
                "ln": f"\\ln\\left({input_var}\\right)",
                "linear_ln": f"\\ln\\left({c[0]} {input_var} \\right)",
                "cos": f"\\cos\\left({input_var}\\right)",
                "linear_cos": f"\\cos\\left({c[0]} {input_var} \\right)",
                "sin": f"\\sin\\left({input_var}\\right)",
                "linear_sin": f"\\sin\\left({c[0]} {input_var} \\right)",
                "tanh": f"\\tanh\\left({input_var}\\right)",
                "linear_tanh": f"\\tanh\\left({c[0]} {input_var} \\right)",
                "classifier": classifier_str,
            }
    else:  # with bias
        if output_format == "console":
            labels = {
                "none": "",
                "add": f"+ {input_var}",
                "subtract": f"- {input_var}",
                "mult": f"{c[0]} * {input_var}",
                "linear": f"{c[0]} * {input_var} + {c[1]}",
                "relu": f"ReLU({input_var})",
                "linear_relu": f"ReLU({c[0]} * {input_var} + {c[1]} )",
                "logistic": f"logistic({input_var})",
                "linear_logistic": f"logistic({c[0]} * {input_var} + {c[1]})",
                "exp": f"exp({input_var})",
                "linear_exp": f"exp({c[0]} * {input_var} + {c[1]})",
                "reciprocal": f"1 / {input_var}",
                "linear_reciprocal": f"1 / ({c[0]} * {input_var} + {c[1]})",
                "ln": f"ln({input_var})",
                "linear_ln": f"ln({c[0]} * {input_var} + {c[1]})",
                "cos": f"cos({input_var})",
                "linear_cos": f"cos({c[0]} * {input_var} + {c[1]})",
                "sin": f"sin({input_var})",
                "linear_sin": f"sin({c[0]} * {input_var} + {c[1]})",
                "tanh": f"tanh({input_var})",
                "linear_tanh": f"tanh({c[0]} * {input_var} + {c[1]})",
                "classifier": classifier_str,
            }
        elif output_format == "latex":
            labels = {
                "none": "",
                "add": f"+ {input_var}",
                "subtract": f"- {input_var}",
                "mult": f"{c[0]} * {input_var}",
                "linear": f"{c[0]} {input_var} + {c[1]}",
                "relu": f"\\operatorname{{ReLU}}\\left( {input_var}\\right)",
                "linear_relu": f"\\operatorname{{ReLU}}\\left({c[0]}{input_var} + {c[1]} \\right)",
                "logistic": f"\\sigma\\left( {input_var} \\right)",
                "linear_logistic": f"\\sigma\\left( {c[0]} {input_var} + {c[1]} \\right)",
                "exp": f"e^{input_var}",
                "linear_exp": f"e^{{ {c[0]} {input_var} + {c[1]} }}",
                "reciprocal": f"\\frac{{1}}{{{input_var}}}",
                "linear_reciprocal": f"\\frac{{1}} {{ {c[0]}{input_var} + {c[1]} }}",
                "ln": f"\\ln\\left({input_var}\\right)",
                "linear_ln": f"\\ln\\left({c[0]} {input_var} + {c[1]} \\right)",
                "cos": f"\\cos\\left({input_var}\\right)",
                "linear_cos": f"\\cos\\left({c[0]} {input_var} + {c[1]} \\right)",
                "sin": f"\\sin\\left({input_var}\\right)",
                "linear_sin": f"\\sin\\left({c[0]} {input_var} + {c[1]} \\right)",
                "tanh": f"\\tanh\\left({input_var}\\right)",
                "linear_tanh": f"\\tanh\\left({c[0]} {input_var} + {c[1]} \\right)",
                "classifier": classifier_str,
            }

    if op_name not in labels:
        raise NotImplementedError(
            f"operation '{op_name}' is not defined for output_format '{output_format}'"
        )

    return labels[op_name]


class Identity(nn.Module):
    """
    A pytorch module implementing the identity function.

    $$
    x = x
    $$
    """

    def __init__(self):
        """
        Initializes the identify function.
        """
        super(Identity, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the identity function.

        Arguments:
            x: input tensor
        """
        return x


class NegIdentity(nn.Module):
    """
    A pytorch module implementing the inverse of an identity function.

    $$
    x = -x
    $$
    """

    def __init__(self):
        """
        Initializes the inverse of an identity function.
        """
        super(NegIdentity, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the inverse of an identity function.

        Arguments:
            x: input tensor
        """
        return -x


class Exponential(nn.Module):
    """
    A pytorch module implementing the exponential function.

    $$
    x = e^x
    $$
    """

    def __init__(self):
        """
        Initializes the exponential function.
        """
        super(Exponential, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the exponential function.

        Arguments:
            x: input tensor
        """
        return torch.exp(x)


class Cosine(nn.Module):
    r"""
    A pytorch module implementing the cosine function.

    $$
    x = \cos(x)
    $$
    """

    def __init__(self):
        """
        Initializes the cosine function.
        """
        super(Cosine, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the cosine function.

        Arguments:
            x: input tensor
        """
        return torch.cos(x)


class Sine(nn.Module):
    r"""
    A pytorch module implementing the sine function.

    $$
    x = \sin(x)
    $$
    """

    def __init__(self):
        """
        Initializes the sine function.
        """
        super(Sine, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the sine function.

        Arguments:
            x: input tensor
        """
        return torch.sin(x)


class Tangens_Hyperbolicus(nn.Module):
    r"""
    A pytorch module implementing the tangens hyperbolicus function.

    $$
    x = \tanh(x)
    $$
    """

    def __init__(self):
        """
        Initializes the tangens hyperbolicus function.
        """
        super(Tangens_Hyperbolicus, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the tangens hyperbolicus function.

        Arguments:
            x: input tensor
        """
        return torch.tanh(x)


class NatLogarithm(nn.Module):
    r"""
    A pytorch module implementing the natural logarithm function.

    $$
    x = \ln(x)
    $$

    """

    def __init__(self):
        """
        Initializes the natural logarithm function.
        """
        super(NatLogarithm, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the natural logarithm function.

        Arguments:
            x: input tensor
        """
        # make sure x is in domain of natural logarithm
        mask = x.clone()
        mask[(x <= 0.0).detach()] = 0
        mask[(x > 0.0).detach()] = 1

        epsilon = 1e-10
        result = torch.log(nn.functional.relu(x) + epsilon) * mask

        return result


class MultInverse(nn.Module):
    r"""
    A pytorch module implementing the multiplicative inverse.

    $$
    x = \frac{1}{x}
    $$
    """

    def __init__(self):
        """
        Initializes the multiplicative inverse.
        """
        super(MultInverse, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the multiplicative inverse.

        Arguments:
            x: input tensor
        """
        return torch.pow(x, -1)


class Zero(nn.Module):
    """
    A pytorch module implementing the zero operation (i.e., a null operation). A zero operation
    presumes that there is no relationship between the input and output.

    $$
    x = 0
    $$
    """

    def __init__(self, stride):
        """
        Initializes the zero operation.
        """
        super(Zero, self).__init__()
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the zero operation.

        Arguments:
            x: input tensor
        """
        if self.stride == 1:
            return x.mul(0.0)
        return x[:, :, :: self.stride, :: self.stride].mul(0.0)


class Softplus(nn.Module):
    r"""
    A pytorch module implementing the softplus function:

    $$
    \operatorname{Softplus}(x) = \frac{1}{β} \operatorname{log} \left( 1 + e^{β x} \right)
    $$
    """

    # This docstring is a raw-string (it starts `r"""` rather than `"""`)
    # so backslashes need not be escaped

    def __init__(self):
        """
        Initializes the softplus function.
        """
        super(Softplus, self).__init__()
        # self.beta = nn.Linear(1, 1, bias=False)
        self.beta = nn.Parameter(torch.ones(1))
        # elf.softplus = nn.Softplus(beta=self.beta)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the softplus function.

        Arguments:
            x: input tensor
        """
        y = torch.log(1 + torch.exp(self.beta * x)) / self.beta
        # y = self.softplus(x)
        return y


class Softminus(nn.Module):
    """
    A pytorch module implementing the softminus function:

    $$
    \\operatorname{Softminus}(x) = x - \\operatorname{log} \\left( 1 + e^{β x} \\right)
    $$
    """

    # This docstring is a normal string, so backslashes need to be escaped

    def __init__(self):
        """
        Initializes the softminus function.
        """
        super(Softminus, self).__init__()
        # self.beta = nn.Linear(1, 1, bias=False)
        self.beta = nn.Parameter(torch.ones(1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the softminus function.

        Arguments:
            x: input tensor
        """
        y = x - torch.log(1 + torch.exp(self.beta * x)) / self.beta
        return y


# defines all the operations. affine is turned off for cuda (optimization prposes)
OPS = {
    "none": Zero(1),
    "add": nn.Sequential(Identity()),
    "subtract": nn.Sequential(NegIdentity()),
    "mult": nn.Sequential(
        nn.Linear(1, 1, bias=False),
    ),
    "linear": nn.Sequential(nn.Linear(1, 1, bias=True)),
    "relu": nn.Sequential(
        nn.ReLU(inplace=False),
    ),
    "linear_relu": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        nn.ReLU(inplace=False),
    ),
    "logistic": nn.Sequential(
        nn.Sigmoid(),
    ),
    "linear_logistic": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        nn.Sigmoid(),
    ),
    "exp": nn.Sequential(
        Exponential(),
    ),
    "linear_exp": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        Exponential(),
    ),
    "cos": nn.Sequential(
        Cosine(),
    ),
    "linear_cos": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        Cosine(),
    ),
    "sin": nn.Sequential(
        Sine(),
    ),
    "linear_sin": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        Sine(),
    ),
    "tanh": nn.Sequential(
        Tangens_Hyperbolicus(),
    ),
    "linear_tanh": nn.Sequential(
        nn.Linear(1, 1, bias=True),
        Tangens_Hyperbolicus(),
    ),
    "reciprocal": nn.Sequential(
        MultInverse(),
    ),
    "linear_reciprocal": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        MultInverse(),
    ),
    "ln": nn.Sequential(
        NatLogarithm(),
    ),
    "linear_ln": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        NatLogarithm(),
    ),
    "softplus": nn.Sequential(
        Softplus(),
    ),
    "linear_softplus": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        Softplus(),
    ),
    "softminus": nn.Sequential(
        Softminus(),
    ),
    "linear_softminus": nn.Sequential(
        nn.Linear(1, 1, bias=False),
        Softminus(),
    ),
}

# this is the list of primitives actually used,
# and it should be a set of names contained in the OPS dictionary
PRIMITIVES = (
    "none",
    "add",
    "subtract",
    "linear",
    "linear_logistic",
    "mult",
    "linear_relu",
)

# make sure that every primitive is in the OPS dictionary
for name in PRIMITIVES:
    assert name in OPS



from .architect import Architect
from .dataset import DARTSDataset, darts_dataset_from_ndarray
from .model_search import DARTSType, Network
from .operations import PRIMITIVES
from .utils import (
    AvgrageMeter,
    format_input_target,
    get_loss_function,
    get_output_format,
    get_output_str,
)
from .visualize import darts_model_plot



from typing import Optional, Tuple

import numpy as np
import torch
from torch.utils.data import Dataset


class DARTSDataset(Dataset):
    """
    A dataset for the DARTS algorithm.
    """

    def __init__(self, input_data: torch.tensor, output_data: torch.tensor):
        """
        Initializes the dataset.

        Arguments:
            input_data: The input data to the dataset.
            output_data: The output data to the dataset.
        """
        assert input_data.shape[0] == output_data.shape[0]
        self.input_data = input_data
        self.output_data = output_data

    def __len__(self, experiment_id: Optional[int] = None) -> int:
        """
        Returns the length of the dataset.

        Arguments:
            experiment_id:

        Returns:
            The length of the dataset.
        """
        return self.input_data.shape[0]

    def __getitem__(self, idx: int) -> Tuple[torch.tensor, torch.tensor]:
        """
        Returns the item at the given index.

        Arguments:
            idx: The index of the item to return.

        Returns:
            The item at the given index.

        """
        input_tensor = self.input_data[idx]
        output_tensor = self.output_data[idx]
        return input_tensor, output_tensor


def darts_dataset_from_ndarray(
    input_data: np.ndarray, output_data: np.ndarray
) -> DARTSDataset:
    """
    A function to create a dataset from numpy arrays.

    Arguments:
        input_data: The input data to the dataset.
        output_data: The output data to the dataset.

    Returns:
        The dataset.

    """

    obj = DARTSDataset(
        torch.tensor(input_data, dtype=torch.get_default_dtype()),
        torch.tensor(output_data, dtype=torch.get_default_dtype()),
    )
    return obj



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autora',
 'autora.skl',
 'autora.theorist',
 'autora.theorist.darts',
 'autora.variable']

package_data = \
{'': ['*']}

install_requires = \
['graphviz>=0.14.1,<0.15.0',
 'imageio>=2.9.0,<3.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18,<2.0',
 'pandas>=1.4.2,<2.0.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.7.1,<2.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'torch>=1.11.0,<2.0.0',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{'tinkerforge': ['tinkerforge>=2.1.25,<3.0.0']}

setup_kwargs = {
    'name': 'autora',
    'version': '1.0.0',
    'description': 'Autonomous Research Assistant (AutoRA) is a framework for implementing machine learning tools which autonomously and iteratively generate 1) new theories to describe real-world data, and 2) experiments to invalidate those theories and seed a new cycle of theory-making. The experiments will be run online via crowd-sourcing platforms (MTurk, Prolific).',
    'long_description': '# Autonomous Research Assistant\nAutonomous Research Assistant (AutoRA) is an open source AI-based system for automating each aspect of empirical research in the behavioral sciences, from the construction of a scientific hypothesis to conducting novel experiments.\n\n# Getting started\n\nYou should be familiar with the command line for your operating system. The topics required are covered in:\n- **macOS**: Joe Kissell. [*Take Control of the Mac Command Line with Terminal, 3rd Edition*]((https://bruknow.library.brown.edu/permalink/01BU_INST/528fgv/cdi_safari_books_v2_9781947282513)). Take Control Books, 2022. Chapters *Read Me First* through *Bring the Command Line Into The Real World*.\n- **Linux**: William E. Shotts. [*The Linux Command Line: a Complete Introduction. 2nd edition.*](https://bruknow.library.brown.edu/permalink/01BU_INST/9mvq88/alma991043239704906966). No Starch Press, 2019. Parts *I: Learning the Shell* and *II: Configuration and the Environment*.\n\nTo use the AutoRA package you need:\n- `python` and packages as specified in the `pyproject.toml` file,\n- `graphviz` for some visualizations.\n\nTo develop the AutoRA package, you also need:\n- `git`, the source control tool,\n- `pre-commit` which is used for handling git pre-commit hooks.\n\nWe recommend setting up your development environment using:\n- `pyenv` which is used for installing different versions of `python`,\n- `poetry`, which handles resolving dependencies between `python` modules and ensures that you are using the same package versions as other members of the development team.\n\nYou should also consider using an IDE. We recommend: \n- PyCharm (academic licenses for PyCharm professional edition are available for free). This is a `python`-specific integrated development environment which comes with extremely powerful tools for changing the structure of `python` code, running tests, etc. \n- Visual Studio Code (free). This is a powerful general text editor with plugins to support `python` development. \n\nThe following sections describe how to install and configure the recommended setup for developing AutoRA.\n\n*Note: For end-users, it may be more appropriate to use an environment manager like `Anaconda` or `Miniconda` instead of `poetry`, but this is not currently supported.*\n\n\n## Development Setup on macOS\n\n### Prerequisites\n\nFor macOS, we strongly recommend using `homebrew` to manage packages.\n\nVisit [https://brew.sh](https://brew.sh) and run the installation instructions.\n\n### Clone Repository\n\nWe recommend using the GitHub CLI to clone the repository. Install it: \n\n```shell\nbrew install gh\n```\n\nClone the repository. Run:\n```shell\ngh repo clone AutoResearch/AutoRA\n```\n\n... and following the prompts to authenticate to GitHub. It should clone the repository to a new directory. This is referred to as the `<project directory>` in the rest of this readme.\n\n### Install Dependencies\n\nOpen the repository directory in the terminal.\n\nInstall the dependencies, which are listed in the [`Brewfile`](./Brewfile) by running:\n\n```shell\nbrew bundle\n```\n\n### Install `python`\n\nWe recommend using `pyenv` to manage `python` versions. \n\n#### Initialize pyenv\nRun the initialization script as follows:\n\n```shell\npyenv init\n``` \n... and follow the instructions to add `pyenv` to the `$PATH`.\n\n#### Restart shell session\n\nAfter making these changes, restart your shell session by executing:\n\n```shell\nexec "$SHELL" \n```\n\n#### Install `python`\n\nInstall a `python` version listed in the [`pyproject.toml`](./pyproject.toml) file. The entry looks like:  \n\n```toml\npython = \'>=3.8.13,<3.11\'\n```\n\nIn this case, you could install version 3.8.13 as follows:\n\n```shell\npyenv install 3.8.13\n```\n\n### Install Pre-Commit Hooks\n\nIf you wish to commit to the repository, you should install the pre-commit hooks with the following command: \n```shell\npre-commit install\n```\n\nFor more information on pre-commit hooks, see [Pre-Commit-Hooks](#pre-commit-hooks)\n\n### Configure your development environment\n\nThere are two suggested options for initializing an environment:\n- _(Recommended)_ Using PyCharm,\n- _(Advanced)_ Using `poetry` from the command line.\n\n#### PyCharm configuration\n\nSet up the Virtual environment – an isolated version of `python` and all the packages required to run AutoRA and develop it further – as follows:\n- Open the `<project directory>` in PyCharm.\n- Navigate to PyCharm > Preferences > Project: AutoRA > Python Interpreter\n- Next to the drop-down list of available interpreters, click the "gear" symbol and choose "Add" to initialize a new interpreter. \n- Select "Poetry environment" in the list on the left. Specify the following:  \n  - Python executable: select the path to the installed `python` version you wish to use, e.g. \n    `~/.pyenv/versions/3.8.13/bin/python3`\n  - Select "install packages from pyproject.toml"\n  - Poetry executable: select the path to the poetry installation you have, e.g. \n    `/opt/homebrew/bin/poetry`\n  - Click "OK" and wait while the environment builds.\n  - Run the "Python tests in tests/" Run/Debug configuration in the PyCharm interface, and check that there are no errors.\n\nAdditional setup steps for PyCharm:\n\n- You can (and should) completely hide the IDE-specific directory for Visual Studio Code in PyCharm by adding `.vscode` to the list of ignored folder names in Preferences > Editor > File Types > Ignored Files and Folders. This only needs to be done once.\n    \n#### Command Line `poetry` Setup\n\nIf you need more control over the `poetry` environment, then you can set up a new environment from the command line.\n\n*Note: Setting up a `poetry` environment on the command line is the only option for VSCode users.*\n\nFrom the `<project directory>`, run the following commands.\n\nActivate the target version of `python` using `pyenv`:\n```shell\npyenv shell 3.8.13\n```\n\nSet up a new poetry environment with that `python` version:\n```shell\npoetry env use $(pyenv which python) \n```\n\nUpdate the installation utilities within the new environment:\n```shell\npoetry run python -m pip install --upgrade pip setuptools wheel\n```\n\nUse the `pyproject.toml` file to resolve and then install all the dependencies\n```shell\npoetry install\n```\n\nOnce this step has been completed, skip to the section [Activating and using the environment](#activating-and-using-the-environment) to test it.\n\n#### Visual Studio Code Configuration\n\nAfter installing Visual Studio Code and the other prerequisites, carry out the following steps:\n\n- Open the `<project directory>` in Visual Studio Code\n- Install the Visual Studio Code plugin recommendations suggested with the project. These include:\n  - `python`\n  - `python-environment-manager`\n- Run the [Command Line poetry Setup](#command-line-poetry-setup) specified above. This can be done in the built-in terminal if desired (Menu: Terminal > New Terminal).\n- Select the `python` option in the vertical bar on the far left of the window (which appear after installing the plugins). Under the title "PYTHON: ENVIRONMENTS" should be a list of `python` environments. If these do not appear:\n  - Refresh the window pane\n  - Ensure the python-environment-manager is installed correctly.\n  - Ensure the python-environment-manager is activated.\n\n- Locate the correct `poetry` environment. Click the "thumbs up" symbol next to the poetry environment name to "set as active workspace interpreter".\n\n- Check that the `poetry` environment is correctly set-up. \n  - Open a new terminal within Visual Studio Code (Menu: Terminal > New Terminal). \n  - It should execute something like `source /Users/me/Library/Caches/pypoetry/virtualenvs/autora-2PgcgopX-py3.8/bin/activate` before offering you a prompt.\n  - If you execute `which python` it should return the path to your python executable in the `.../autora-2PgcgopX-py3.8/bin` directory.\n  - Ensure that there are no errors when you run: \n    ```shell\n    python -m unittest\n    ```\n    in the built-in terminal. \n\n### Activating and using the environment\n\n#### Using `poetry` interactively\n\nTo run interactive commands, you can activate the poetry virtual environment. From the `<project directory>` directory, run:\n\n```shell\npoetry shell\n```\n\nThis spawns a new shell where you have access to the poetry `python` and all the packages installed using `poetry install`. You should see the prompt change:\n\n```\n% poetry shell\nSpawning shell within /Users/me/Library/Caches/pypoetry/virtualenvs/autora-2PgcgopX-py3.8\nRestored session: Fri Jun 24 12:34:56 EDT 2022\n(autora-2PgcgopX-py3.8) % \n```\n\nIf you execute `python` and then `import numpy`, you should be able to see that `numpy` has been imported from the `autora-2PgcgopX-py3.8` environment:\n\n```\n(autora-2PgcgopX-py3.8) % python\nPython 3.8.13 (default, Jun 16 2022, 12:34:56) \n[Clang 13.1.6 (clang-1316.0.21.2.5)] on darwin\nType "help", "copyright", "credits" or "license" for more information.\n>>> import numpy\n>>> numpy\n<module \'numpy\' from \'/Users/me/Library/Caches/pypoetry/virtualenvs/autora-2PgcgopX-py3.8/lib/python3.8/site-packages/numpy/__init__.py\'>\n```\n\nTo deactivate the `poetry` environment, `exit` the session. This should return you to your original prompt, as follows:\n```\n(autora-2PgcgopX-py3.8) % exit\n\nSaving session...\n...saving history...truncating history files...\n...completed.\n% \n```\n\nTo run a script, e.g. the `run_weber_study.py` script in the [`example/weber`](./example/weber) directory, execute: \n\n```shell\npoetry run python example/weber/run_weber_study.py\n```\n\n#### Using `poetry` non-interactively\n\nYou can run python programs using poetry without activating the poetry environment, by using `poetry run {command}`. For example, to run the tests, execute:\n\n```shell\npoetry run python -m unittest\n```\n\nIt should return something like:\n\n```\n% poetry run python -m unittest\n.\n--------------------------------\nRan 1 test in 0.000s\n\nOK\n```\n\n## Development Setup on Windows\n\nWindows is not yet officially supported. You may be able to follow the same approach as for macOS to set up your development environment, with some modifications, e.g.:\n- Using `chocolatey` in place of `homebrew`,\n- Using the GitHub Desktop application in place of the GitHub CLI.\n\nIf you successfully set up AutoRA on Windows, please update this readme.\n\n## Development Practices\n\n### Pre-Commit Hooks\n\nWe use [`pre-commit`](https://pre-commit.com) to manage pre-commit hooks. \n\nPre-commit hooks are programs which run before each git commit, and can read and potentially modify the files which are to be committed. \n\nWe use pre-commit hooks to:\n- enforce coding guidelines, including the `python` style-guide [PEP8](https://peps.python.org/pep-0008/) (`black` and `flake8`), \n- to check the order of `import` statements (`isort`),\n- to check the types of `python` objects (`mypy`).\n\nThe hooks and their settings are specified in [`.pre-commit-config.yaml`](./.pre-commit-config.yaml).\n\nSee the section [Install Pre-commit Hooks](#install-pre-commit-hooks) for installation instructions.\n\n#### Handling Pre-Commit Hook Errors\n\nIf your `git commit` fails because of the pre-commit hook, then you should:\n\n1. Run the pre-commit hooks on the files which you have staged, by running the following command in your terminal: \n    ```zsh\n    $ pre-commit run\n    ```\n\n2. Inspect the output. It might look like this:\n   ```\n   $ pre-commit run\n   black....................Passed\n   isort....................Passed\n   flake8...................Passed\n   mypy.....................Failed\n   - hook id: mypy\n   - exit code: 1\n   \n   example.py:33: error: Need type annotation for "data" (hint: "data: Dict[<type>, <type>] = ...")\n   Found 1 errors in 1 files (checked 10 source files)\n   ```\n3. Fix any errors which are reported.\n   **Important: Once you\'ve changed the code, re-stage the files it to Git. \n   This might mean unstaging changes and then adding them again.**\n4. If you have trouble:\n   - Do a web-search to see if someone else had a similar error in the past.\n   - Check that the tests you\'ve written work correctly.\n   - Check that there aren\'t any other obvious errors with the code.\n   - If you\'ve done all of that, and you still can\'t fix the problem, get help from someone else on the team.\n5. Repeat 1-4 until all hooks return "passed", e.g.\n   ```\n   $ pre-commit run\n   black....................Passed\n   isort....................Passed\n   flake8...................Passed\n   mypy.....................Passed\n   ```\n\nIt\'s easiest to solve these kinds of problems if you make small commits, often.  \n\n# Documentation\n\n## Commenting code\n\nTo help users understand code better, and to make the documentation generation automatic, we have some standards for documenting code. The comments, docstrings, and the structure of the code itself are meant to make life easier for the reader. \n- If something important isn\'t _obvious_ from the code, then it should be _made_ obvious with a comment. \n- Conversely, if something _is_ obvious, then it doesn\'t need a comment.\n\nThese standards are inspired by John Ousterhout. *A Philosophy of Software Design.* Yaknyam Press, 2021. Chapter 12 – 14.\n\n### Every public function, class and method has documentation\n\nWe include docstrings for all public functions, classes, and methods. These docstrings are meant to give a concise, high-level overview of **why** the function exists, **what** it is trying to do, and what is **important** about the code. (Details about **how** the code works are often better placed in detailed comments within the code.)\n\nEvery function, class or method has a one-line **high-level description** which clarifies its intent.   \n\nThe **meaning** and **type** of all the input and output parameters should be described.\n\nThere should be **examples** of how to use the function, class or method, with expected outputs, formatted as ["doctests"](https://docs.python.org/3/library/doctest.html). These should include normal cases for the function, but also include cases where it behaves unexpectedly or fails. \n\nWe follow the [Google Style Python Docstrings](https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html), as these are supported by the online documentation tool we use (see [Online Documentation](#online-documentation)).\n\nA well documented function looks something like this:\n```python\ndef first_order_linear(\n    x: Union[float, np.ndarray], c: float, m: float\n) -> Union[float, np.ndarray]:\n    """\n    Evaluate a first order linear model of the form y = m x + c.\n\n    Arguments:\n        x: input location(s) on the x-axis\n        c: y-intercept of the linear model\n        m: gradient of the linear model\n\n    Returns:\n        y: result y = m x + c, the same shape and type as x\n\n    Examples:\n        >>> first_order_linear(0. , 1. , 0. )\n        1.0\n        >>> first_order_linear(np.array([-1. , 0. , 1. ]), c=1.0, m=2.0)\n        array([-1.,  1.,  3.])\n    """\n    y = m * x + c\n    return y\n```\n\nFor an example of a file with docstrings, see [docs/docstrings.py](./docs/docstrings.py).\n\n*Pro-Tip: Write the docstring for your new high-level object before starting on the code. In particular, writing examples of how you expect it should be used can help clarify the right level of abstraction.*\n\n## Online Documentation\n\nOnline Documentation is automatically generated using [Material for MkDocs](https://squidfunk.github.io/mkdocs-material/) based on docstrings in files in the `autora/` directory. \n\n### Commands\n\nBuild and serve the documentation using the following commands:\n\n* `poetry run mkdocs serve` - Start the live-reloading docs server.\n* `poetry run mkdocs build` - Build the documentation site.\n* `poetry run mkdocs gh-deploy` - Build the documentation and serve at https://AutoResearch.github.io/AutoRA/\n* `poetry run mkdocs -h` - Print help message and exit.\n\n### Documentation layout\n```\nmkdocs.yml    # The configuration file for the documentation.\ndocs/         # Directory for static pages to be included in the documentation.\n    index.md  # The documentation homepage.\n    ...       # Other markdown pages, images and other files.\nautora/          # The directory containing the source code.\n```\n\n\n',
    'author': 'Sebastian Musslick',
    'author_email': 'sebastian_musslick@brown.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://musslick.github.io/AER_website/Research.html',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import torch
import torch.nn as nn


class Fan_Out(nn.Module):
    """
    A neural network class that splits a given input vector into separate nodes. Each element of
    the original input vector is allocated a separate node in a computation graph.
    """

    def __init__(self, num_inputs: int):
        """
        Initialize the Fan Out operation.

        Arguments:
                num_inputs (int): The number of distinct input nodes to generate
        """
        super(Fan_Out, self).__init__()

        self._num_inputs = num_inputs

        self.input_output = list()
        for i in range(num_inputs):
            linearConnection = nn.Linear(num_inputs, 1, bias=False)
            linearConnection.weight.data = torch.zeros(1, num_inputs)
            linearConnection.weight.data[0, i] = 1
            linearConnection.weight.requires_grad = False
            self.input_output.append(linearConnection)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Fan Out operation.

        Arguments:
            input: input vector whose elements are split into separate input nodes
        """

        output = list()
        for i in range(self._num_inputs):
            output.append(self.input_output[i](input))

        return output
